from __future__ import absolute_import, unicode_literals

__all__ = ['StandardCollection', 'VertexCollection', 'EdgeCollection']

from numbers import Number

from .api import APIWrapper
from .cursor import Cursor
from .exceptions import (
    CollectionChecksumError,
    CollectionConfigureError,
    CollectionLoadError,
    CollectionPropertiesError,
    CollectionRecalculateCountError,
    CollectionResponsibleShardError,
    CollectionRenameError,
    CollectionRevisionError,
    CollectionRotateJournalError,
    CollectionStatisticsError,
    CollectionTruncateError,
    CollectionUnloadError,
    DocumentCountError,
    DocumentInError,
    DocumentDeleteError,
    DocumentGetError,
    DocumentKeysError,
    DocumentIDsError,
    DocumentInsertError,
    DocumentParseError,
    DocumentReplaceError,
    DocumentRevisionError,
    DocumentUpdateError,
    EdgeListError,
    IndexCreateError,
    IndexDeleteError,
    IndexListError,
    IndexLoadError,
)
from .formatter import (
    format_collection,
    format_edge,
    format_index,
    format_vertex,
)
from .request import Request
from .utils import (
    get_doc_id,
    is_none_or_int,
    is_none_or_str,
)


class Collection(APIWrapper):
    """Base class for collection API wrappers.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    :param name: Collection name.
    :type name: str | unicode
    """

    types = {
        2: 'document',
        3: 'edge'
    }

    statuses = {
        1: 'new',
        2: 'unloaded',
        3: 'loaded',
        4: 'unloading',
        5: 'deleted',
        6: 'loading'
    }

    def __init__(self, connection, executor, name):
        super(Collection, self).__init__(connection, executor)
        self._name = name
        self._id_prefix = name + '/'

    # def __contains__(self, document):
    #     return self.has(document, check_rev=False)

    def _get_status_text(self, code):  # pragma: no cover
        """Return the collection status text.

        :param code: Collection status code.
        :type code: int
        :return: Collection status text or None if code is None.
        :rtype: str | unicode
        """
        return None if code is None else self.statuses[code]

    def _validate_id(self, doc_id):
        """Check the collection name in the document ID.

        :param doc_id: Document ID.
        :type doc_id: str | unicode
        :return: Verified document ID.
        :rtype: str | unicode
        :raise arango.exceptions.DocumentParseError: On bad collection name.
        """
        if not doc_id.startswith(self._id_prefix):
            raise DocumentParseError(
                'bad collection name in document ID "{}"'.format(doc_id))
        return doc_id

    def _extract_id(self, body):
        """Extract the document ID from document body.

        :param body: Document body.
        :type body: dict
        :return: Document ID.
        :rtype: str | unicode
        :raise arango.exceptions.DocumentParseError: On missing ID and key.
        """
        try:
            if '_id' in body:
                return self._validate_id(body['_id'])
            else:
                return self._id_prefix + body['_key']
        except KeyError:
            raise DocumentParseError('field "_key" or "_id" required')

    def _prep_from_body(self, document, check_rev):
        """Prepare document ID and request headers.

        :param document: Document body.
        :type document: dict
        :param check_rev: Whether to check the revision.
        :type check_rev: bool
        :return: Document ID and request headers.
        :rtype: (str | unicode, dict)
        """
        doc_id = self._extract_id(document)
        if not check_rev or '_rev' not in document:
            return doc_id, {}
        return doc_id, {'If-Match': document['_rev']}

    def _prep_from_doc(self, document, rev, check_rev):
        """Prepare document ID, body and request headers.

        :param document: Document ID, key or body.
        :type document: str | unicode | dict
        :param rev: Document revision or None.
        :type rev: str | unicode | None
        :param check_rev: Whether to check the revision.
        :type check_rev: bool
        :return: Document ID, body and request headers.
        :rtype: (str | unicode, str | unicode | body, dict)
        """
        if isinstance(document, dict):
            doc_id = self._extract_id(document)
            rev = rev or document.get('_rev')

            if not check_rev or rev is None:
                return doc_id, doc_id, {}
            else:
                return doc_id, doc_id, {'If-Match': rev}
        else:
            if '/' in document:
                doc_id = self._validate_id(document)
            else:
                doc_id = self._id_prefix + document

            if not check_rev or rev is None:
                return doc_id, doc_id, {}
            else:
                return doc_id, doc_id, {'If-Match': rev}

    def _ensure_key_in_body(self, body):
        """Return the document body with "_key" field populated.

        :param body: Document body.
        :type body: dict
        :return: Document body with "_key" field.
        :rtype: dict
        :raise arango.exceptions.DocumentParseError: On missing ID and key.
        """
        if '_key' in body:
            return body
        elif '_id' in body:
            doc_id = self._validate_id(body['_id'])
            body = body.copy()
            body['_key'] = doc_id[len(self._id_prefix):]
            return body
        raise DocumentParseError('field "_key" or "_id" required')

    def _ensure_key_from_id(self, body):
        """Return the body with "_key" field if it has "_id" field.

        :param body: Document body.
        :type body: dict
        :return: Document body with "_key" field if it has "_id" field.
        :rtype: dict
        """
        if '_id' in body and '_key' not in body:
            doc_id = self._validate_id(body['_id'])
            body = body.copy()
            body['_key'] = doc_id[len(self._id_prefix):]
        return body

    @property
    def name(self):
        """Return collection name.

        :return: Collection name.
        :rtype: str | unicode
        """
        return self._name

    async def recalculate_count(self):
        """Recalculate the document count.

        :return: True if recalculation was successful.
        :rtype: bool
        :raise arango.exceptions.CollectionRecalculateCountError: If operation
            fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/recalculateCount'.format(self.name)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionRecalculateCountError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def responsible_shard(self, document):  # pragma: no cover
        """Return the ID of the shard responsible for given **document**.

        If the document does not exist, return the shard that would be
        responsible.

        :return: Shard ID
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/responsibleShard'.format(self.name),
            data=document,
            read=self.name
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body['shardId']
            raise CollectionResponsibleShardError(resp, request)

        return await self._execute(request, response_handler)

    async def rename(self, new_name):
        """Rename the collection.

        Renames may not be reflected immediately in async execution, batch
        execution or transactions. It is recommended to initialize new API
        wrappers after a rename.

        :param new_name: New collection name.
        :type new_name: str | unicode
        :return: True if collection was renamed successfully.
        :rtype: bool
        :raise arango.exceptions.CollectionRenameError: If rename fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/rename'.format(self.name),
            data={'name': new_name}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionRenameError(resp, request)
            self._name = new_name
            self._id_prefix = new_name + '/'
            return True

        return await self._execute(request, response_handler)

    async def properties(self):
        """Return collection properties.

        :return: Collection properties.
        :rtype: dict
        :raise arango.exceptions.CollectionPropertiesError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/collection/{}/properties'.format(self.name),
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionPropertiesError(resp, request)
            return format_collection(resp.body)

        return await self._execute(request, response_handler)

    async def configure(self, sync=None, journal_size=None):
        """Configure collection properties.

        :param sync: Block until operations are synchronized to disk.
        :type sync: bool
        :param journal_size: Journal size in bytes.
        :type journal_size: int
        :return: New collection properties.
        :rtype: dict
        :raise arango.exceptions.CollectionConfigureError: If operation fails.
        """
        data = {}
        if sync is not None:
            data['waitForSync'] = sync
        if journal_size is not None:
            data['journalSize'] = journal_size

        request = Request(
            method='put',
            endpoint='/_api/collection/{}/properties'.format(self.name),
            data=data
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionConfigureError(resp, request)
            return format_collection(resp.body)

        return await self._execute(request, response_handler)

    async def statistics(self):
        """Return collection statistics.

        :return: Collection statistics.
        :rtype: dict
        :raise arango.exceptions.CollectionStatisticsError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/collection/{}/figures'.format(self.name),
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionStatisticsError(resp, request)

            stats = resp.body.get('figures', resp.body)
            for f in ['compactors', 'datafiles', 'journals']:
                if f in stats and 'fileSize' in stats[f]:  # pragma: no cover
                    stats[f]['file_size'] = stats[f].pop('fileSize')
            if 'compactionStatus' in stats:  # pragma: no cover
                status = stats.pop('compactionStatus')
                if 'bytesRead' in status:
                    status['bytes_read'] = status.pop('bytesRead')
                if 'bytesWritten' in status:
                    status['bytes_written'] = status.pop('bytesWritten')
                if 'filesCombined' in status:
                    status['files_combined'] = status.pop('filesCombined')
                stats['compaction_status'] = status
            if 'documentReferences' in stats:  # pragma: no cover
                stats['document_refs'] = stats.pop('documentReferences')
            if 'lastTick' in stats:  # pragma: no cover
                stats['last_tick'] = stats.pop('lastTick')
            if 'waitingFor' in stats:  # pragma: no cover
                stats['waiting_for'] = stats.pop('waitingFor')
            if 'documentsSize' in stats:  # pragma: no cover
                stats['documents_size'] = stats.pop('documentsSize')
            if 'cacheInUse' in stats:  # pragma: no cover
                stats['cache_in_use'] = stats.pop('cacheInUse')
            if 'cacheSize' in stats:  # pragma: no cover
                stats['cache_size'] = stats.pop('cacheSize')
            if 'cacheUsage' in stats:  # pragma: no cover
                stats['cache_usage'] = stats.pop('cacheUsage')
            if 'uncollectedLogfileEntries' in stats:  # pragma: no cover
                stats['uncollected_logfile_entries'] = \
                    stats.pop('uncollectedLogfileEntries')
            return stats

        return await self._execute(request, response_handler)

    async def revision(self):
        """Return collection revision.

        :return: Collection revision.
        :rtype: str | unicode
        :raise arango.exceptions.CollectionRevisionError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/collection/{}/revision'.format(self.name),
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionRevisionError(resp, request)
            return resp.body['revision']

        return await self._execute(request, response_handler)

    async def checksum(self, with_rev=False, with_data=False):
        """Return collection checksum.

        :param with_rev: Include document revisions in checksum calculation.
        :type with_rev: bool
        :param with_data: Include document data in checksum calculation.
        :type with_data: bool
        :return: Collection checksum.
        :rtype: str | unicode
        :raise arango.exceptions.CollectionChecksumError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/collection/{}/checksum'.format(self.name),
            params={'withRevision': with_rev, 'withData': with_data}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionChecksumError(resp, request)
            return resp.body['checksum']

        return await self._execute(request, response_handler)

    async def load(self):
        """Load the collection into memory.

        :return: True if collection was loaded successfully.
        :rtype: bool
        :raise arango.exceptions.CollectionLoadError: If operation fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/load'.format(self.name)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionLoadError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def unload(self):
        """Unload the collection from memory.

        :return: True if collection was unloaded successfully.
        :rtype: bool
        :raise arango.exceptions.CollectionUnloadError: If operation fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/unload'.format(self.name)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionUnloadError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def rotate(self):
        """Rotate the collection journal.

        :return: True if collection journal was rotated successfully.
        :rtype: bool
        :raise arango.exceptions.CollectionRotateJournalError: If rotate fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/rotate'.format(self.name),
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionRotateJournalError(resp, request)
            return True  # pragma: no cover

        return await self._execute(request, response_handler)

    async def truncate(self):
        """Delete all documents in the collection.

        :return: True if collection was truncated successfully.
        :rtype: bool
        :raise arango.exceptions.CollectionTruncateError: If operation fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/truncate'.format(self.name)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise CollectionTruncateError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def count(self):
        """Return the total document count.

        :return: Total document count.
        :rtype: int
        :raise arango.exceptions.DocumentCountError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/collection/{}/count'.format(self.name)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentCountError(resp, request)
            return resp.body['count']

        return await self._execute(request, response_handler)

    async def has(self, document, rev=None, check_rev=True):
        """Check if a document exists in the collection.

        :param document: Document ID, key or body. Document body must contain
            the "_id" or "_key" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides value of "_rev" field
            in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :return: True if document exists, False otherwise.
        :rtype: bool
        :raise arango.exceptions.DocumentInError: If check fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, body, headers = self._prep_from_doc(document, rev, check_rev)

        request = Request(
            method='get',
            endpoint='/_api/document/{}'.format(handle),
            headers=headers,
            read=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202:
                return False
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentInError(resp, request)
            return bool(resp.body)

        return await self._execute(request, response_handler)

    async def ids(self):
        """Return the IDs of all documents in the collection.

        :return: Document ID cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.DocumentIDsError: If retrieval fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/simple/all-keys',
            data={'collection': self.name, 'type': 'id'},
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentIDsError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def keys(self):
        """Return the keys of all documents in the collection.

        :return: Document key cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.DocumentKeysError: If retrieval fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/simple/all-keys',
            data={'collection': self.name, 'type': 'key'},
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentKeysError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def all(self, skip=None, limit=None):
        """Return all documents in the collection.

        :param skip: Number of documents to skip.
        :type skip: int
        :param limit: Max number of documents returned.
        :type limit: int
        :return: Document cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert is_none_or_int(skip), 'skip must be a non-negative int'
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        data = {'collection': self.name}
        if skip is not None:
            data['skip'] = skip
        if limit is not None:
            data['limit'] = limit

        request = Request(
            method='put',
            endpoint='/_api/simple/all',
            data=data,
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def export(
            self,
            limit=None,
            count=False,
            batch_size=None,
            flush=False,
            flush_wait=None,
            ttl=None,
            filter_fields=None,
            filter_type='include'):
        """Export all documents in the collection using a server cursor.

        :param flush: If set to True, flush the write-ahead log prior to the
            export. If set to False, documents in the write-ahead log during
            the export are not included in the result.
        :type flush: bool
        :param flush_wait: Max wait time in seconds for write-ahead log flush.
        :type flush_wait: int
        :param count: Include the document count in the server cursor.
        :type count: bool
        :param batch_size: Max number of documents in the batch fetched by
            the cursor in one round trip.
        :type batch_size: int
        :param limit: Max number of documents fetched by the cursor.
        :type limit: int
        :param ttl: Time-to-live for the cursor on the server.
        :type ttl: int
        :param filter_fields: Document fields to filter with.
        :type filter_fields: [str | unicode]
        :param filter_type: Allowed values are "include" or "exclude".
        :type filter_type: str | unicode
        :return: Document cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.DocumentGetError: If export fails.
        """
        data = {'count': count, 'flush': flush}
        if flush_wait is not None:
            data['flushWait'] = flush_wait
        if batch_size is not None:
            data['batchSize'] = batch_size
        if limit is not None:
            data['limit'] = limit
        if ttl is not None:
            data['ttl'] = ttl
        if filter_fields is not None:
            data['restrict'] = {
                'fields': filter_fields,
                'type': filter_type
            }
        request = Request(
            method='post',
            endpoint='/_api/export',
            params={'collection': self.name},
            data=data
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body, 'export')

        return await self._execute(request, response_handler)

    async def find(self, filters, skip=None, limit=None):
        """Return all documents that match the given filters.

        :param filters: Document filters.
        :type filters: dict
        :param skip: Number of documents to skip.
        :type skip: int
        :param limit: Max number of documents returned.
        :type limit: int
        :return: Document cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert isinstance(filters, dict), 'filters must be a dict'
        assert is_none_or_int(skip), 'skip must be a non-negative int'
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        data = {
            'collection': self.name,
            'example': filters,
            'skip': skip,
        }
        if limit is not None:
            data['limit'] = limit

        request = Request(
            method='put',
            endpoint='/_api/simple/by-example',
            data=data,
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def find_near(self, latitude, longitude, limit=None):
        """Return documents near a given coordinate.

        Documents returned are sorted according to distance, with the nearest
        document being the first. If there are documents of equal distance,
        they are randomly chosen from the set until the limit is reached. A geo
        index must be defined in the collection to use this method.

        :param latitude: Latitude.
        :type latitude: int | float
        :param longitude: Longitude.
        :type longitude: int | float
        :param limit: Max number of documents returned.
        :type limit: int
        :returns: Document cursor.
        :rtype: arango.cursor.Cursor
        :raises arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert isinstance(latitude, Number), 'latitude must be a number'
        assert isinstance(longitude, Number), 'longitude must be a number'
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        query = """
        FOR doc IN NEAR(@collection, @latitude, @longitude{})
            RETURN doc
        """.format('' if limit is None else ', @limit ')

        bind_vars = {
            'collection': self._name,
            'latitude': latitude,
            'longitude': longitude
        }
        if limit is not None:
            bind_vars['limit'] = limit

        request = Request(
            method='post',
            endpoint='/_api/cursor',
            data={
                'query': query,
                'bindVars': bind_vars,
                'count': True
            },
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def find_in_range(
            self,
            field,
            lower,
            upper,
            skip=None,
            limit=None):
        """Return documents within a given range in a random order.

        A skiplist index must be defined in the collection to use this method.

        :param field: Document field name.
        :type field: str | unicode
        :param lower: Lower bound (inclusive).
        :type lower: int
        :param upper: Upper bound (exclusive).
        :type upper: int
        :param skip: Number of documents to skip.
        :type skip: int
        :param limit: Max number of documents returned.
        :type limit: int
        :returns: Document cursor.
        :rtype: arango.cursor.Cursor
        :raises arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert is_none_or_int(skip), 'skip must be a non-negative int'
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        bind_vars = {
            '@collection': self._name,
            'field': field,
            'lower': lower,
            'upper': upper,
            'skip': 0 if skip is None else skip,
            'limit': 2147483647 if limit is None else limit,  # 2 ^ 31 - 1
        }

        query = """
        FOR doc IN @@collection
            FILTER doc.@field >= @lower && doc.@field < @upper
            LIMIT @skip, @limit
            RETURN doc
        """

        request = Request(
            method='post',
            endpoint='/_api/cursor',
            data={
                'query': query,
                'bindVars': bind_vars,
                'count': True
            },
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def find_in_radius(self, latitude, longitude, radius, distance_field=None):
        """Return documents within a given radius around a coordinate.

        A geo index must be defined in the collection to use this method.

        :param latitude: Latitude.
        :type latitude: int | float
        :param longitude: Longitude.
        :type longitude: int | float
        :param radius: Max radius.
        :type radius: int | float
        :param distance_field: Document field used to indicate the distance to
            the given coordinate. This parameter is ignored in transactions.
        :type distance_field: str | unicode
        :returns: Document cursor.
        :rtype: arango.cursor.Cursor
        :raises arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert isinstance(latitude, Number), 'latitude must be a number'
        assert isinstance(longitude, Number), 'longitude must be a number'
        assert isinstance(radius, Number), 'radius must be a number'
        assert is_none_or_str(distance_field), 'distance_field must be a str'

        query = """
        FOR doc IN WITHIN(@@collection, @latitude, @longitude, @radius{})
            RETURN doc
        """.format('' if distance_field is None else ', @distance')

        bind_vars = {
            '@collection': self._name,
            'latitude': latitude,
            'longitude': longitude,
            'radius': radius
        }
        if distance_field is not None:
            bind_vars['distance'] = distance_field

        request = Request(
            method='post',
            endpoint='/_api/cursor',
            data={
                'query': query,
                'bindVars': bind_vars,
                'count': True
            },
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def find_in_box(
            self,
            latitude1,
            longitude1,
            latitude2,
            longitude2,
            skip=None,
            limit=None,
            index=None):
        """Return all documents in an rectangular area.

        :param latitude1: First latitude.
        :type latitude1: int | float
        :param longitude1: First longitude.
        :type longitude1: int | float
        :param latitude2: Second latitude.
        :type latitude2: int | float
        :param longitude2: Second longitude
        :type longitude2: int | float
        :param skip: Number of documents to skip.
        :type skip: int
        :param limit: Max number of documents returned.
        :type limit: int
        :param index: ID of the geo index to use (without the collection
            prefix). This parameter is ignored in transactions.
        :type index: str | unicode
        :returns: Document cursor.
        :rtype: arango.cursor.Cursor
        :raises arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert isinstance(latitude1, Number), 'latitude1 must be a number'
        assert isinstance(longitude1, Number), 'longitude1 must be a number'
        assert isinstance(latitude2, Number), 'latitude2 must be a number'
        assert isinstance(longitude2, Number), 'longitude2 must be a number'
        assert is_none_or_int(skip), 'skip must be a non-negative int'
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        data = {
            'collection': self._name,
            'latitude1': latitude1,
            'longitude1': longitude1,
            'latitude2': latitude2,
            'longitude2': longitude2,
        }
        if skip is not None:
            data['skip'] = skip
        if limit is not None:
            data['limit'] = limit
        if index is not None:
            data['geo'] = self._name + '/' + index

        request = Request(
            method='put',
            endpoint='/_api/simple/within-rectangle',
            data=data,
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def find_by_text(self, field, query, limit=None):
        """Return documents that match the given fulltext query.

        :param field: Document field with fulltext index.
        :type field: str | unicode
        :param query: Fulltext query.
        :type query: str | unicode
        :param limit: Max number of documents returned.
        :type limit: int
        :returns: Document cursor.
        :rtype: arango.cursor.Cursor
        :raises arango.exceptions.DocumentGetError: If retrieval fails.
        """
        assert is_none_or_int(limit), 'limit must be a non-negative int'

        bind_vars = {
            'collection': self._name,
            'field': field,
            'query': query,
        }
        if limit is not None:
            bind_vars['limit'] = limit

        aql = """
        FOR doc IN FULLTEXT(@collection, @field, @query{})
            RETURN doc
        """.format('' if limit is None else ', @limit')

        request = Request(
            method='post',
            endpoint='/_api/cursor',
            data={'query': aql, 'bindVars': bind_vars, 'count': True},
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return Cursor(self._conn, resp.body)

        return await self._execute(request, response_handler)

    async def get_many(self, documents):
        """Return multiple documents ignoring any missing ones.

        :param documents: List of document keys, IDs or bodies. Document bodies
            must contain the "_id" or "_key" fields.
        :type documents: [str | unicode | dict]
        :return: Documents. Missing ones are not included.
        :rtype: [dict]
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        """
        handles = [
            self._extract_id(doc) if isinstance(doc, dict) else doc
            for doc in documents
        ]

        request = Request(
            method='put',
            endpoint='/_api/simple/lookup-by-keys',
            data={'collection': self.name, 'keys': handles},
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            docs = resp.body['documents']
            return [doc for doc in docs if '_id' in doc]

        return await self._execute(request, response_handler)

    async def random(self):
        """Return a random document from the collection.

        :return: A random document.
        :rtype: dict
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/simple/any',
            data={'collection': self.name},
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return resp.body['document']

        return await self._execute(request, response_handler)

    ####################
    # Index Management #
    ####################

    async def indexes(self):
        """Return the collection indexes.

        :return: Collection indexes.
        :rtype: [dict]
        :raise arango.exceptions.IndexListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/index',
            params={'collection': self.name},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise IndexListError(resp, request)
            result = resp.body['indexes']
            return [format_index(index) for index in result]

        return await self._execute(request, response_handler)

    async def _add_index(self, data):
        """Helper method for creating a new index.

        :param data: Index data.
        :type data: dict
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/index',
            data=data,
            params={'collection': self.name}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise IndexCreateError(resp, request)
            return format_index(resp.body)

        return await self._execute(request, response_handler)

    async def add_hash_index(
            self,
            fields,
            unique=None,
            sparse=None,
            deduplicate=None,
            name=None,
            in_background=None):
        """Create a new hash index.

        :param fields: Document fields to index.
        :type fields: [str | unicode]
        :param unique: Whether the index is unique.
        :type unique: bool
        :param sparse: If set to True, documents with None in the field
            are also indexed. If set to False, they are skipped.
        :type sparse: bool
        :param deduplicate: If set to True, inserting duplicate index values
            from the same document triggers unique constraint errors.
        :type deduplicate: bool
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'hash', 'fields': fields}

        if unique is not None:
            data['unique'] = unique
        if sparse is not None:
            data['sparse'] = sparse
        if deduplicate is not None:
            data['deduplicate'] = deduplicate
        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def add_skiplist_index(
            self,
            fields,
            unique=None,
            sparse=None,
            deduplicate=None,
            name=None,
            in_background=None):
        """Create a new skiplist index.

        :param fields: Document fields to index.
        :type fields: [str | unicode]
        :param unique: Whether the index is unique.
        :type unique: bool
        :param sparse: If set to True, documents with None in the field
            are also indexed. If set to False, they are skipped.
        :type sparse: bool
        :param deduplicate: If set to True, inserting duplicate index values
            from the same document triggers unique constraint errors.
        :type deduplicate: bool
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'skiplist', 'fields': fields}

        if unique is not None:
            data['unique'] = unique
        if sparse is not None:
            data['sparse'] = sparse
        if deduplicate is not None:
            data['deduplicate'] = deduplicate
        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def add_geo_index(
            self,
            fields,
            ordered=None,
            name=None,
            in_background=None):
        """Create a new geo-spatial index.

        :param fields: A single document field or a list of document fields. If
            a single field is given, the field must have values that are lists
            with at least two floats. Documents with missing fields or invalid
            values are excluded.
        :type fields: str | unicode | list
        :param ordered: Whether the order is longitude, then latitude.
        :type ordered: bool
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'geo', 'fields': fields}

        if ordered is not None:
            data['geoJson'] = ordered
        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def add_fulltext_index(self,
                           fields,
                           min_length=None,
                           name=None,
                           in_background=None):
        """Create a new fulltext index.

        :param fields: Document fields to index.
        :type fields: [str | unicode]
        :param min_length: Minimum number of characters to index.
        :type min_length: int
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'fulltext', 'fields': fields}

        if min_length is not None:
            data['minLength'] = min_length
        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def add_persistent_index(
            self,
            fields,
            unique=None,
            sparse=None,
            name=None,
            in_background=None):
        """Create a new persistent index.

        Unique persistent indexes on non-sharded keys are not supported in a
        cluster.

        :param fields: Document fields to index.
        :type fields: [str | unicode]
        :param unique: Whether the index is unique.
        :type unique: bool
        :param sparse: Exclude documents that do not contain at least one of
            the indexed fields, or documents that have a value of None in any
            of the indexed fields.
        :type sparse: bool
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'persistent', 'fields': fields}

        if unique is not None:
            data['unique'] = unique
        if sparse is not None:
            data['sparse'] = sparse
        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def add_ttl_index(self,
                      fields,
                      expiry_time,
                      name=None,
                      in_background=None):
        """Create a new TTL (time-to-live) index.

        :param fields: Document field to index.
        :type fields: [str | unicode]
        :param expiry_time: Time of expiry in seconds after document creation.
        :type expiry_time: int
        :param name: Optional name for the index.
        :type name: str | unicode
        :param in_background: Do not hold the collection lock.
        :type in_background: bool
        :return: New index details.
        :rtype: dict
        :raise arango.exceptions.IndexCreateError: If create fails.
        """
        data = {'type': 'ttl', 'fields': fields, 'expireAfter': expiry_time}

        if name is not None:
            data['name'] = name
        if in_background is not None:
            data['inBackground'] = in_background

        return await self._add_index(data)

    async def delete_index(self, index_id, ignore_missing=False):
        """Delete an index.

        :param index_id: Index ID.
        :type index_id: str | unicode
        :param ignore_missing: Do not raise an exception on missing index.
        :type ignore_missing: bool
        :return: True if index was deleted successfully, False if index was
            not found and **ignore_missing** was set to True.
        :rtype: bool
        :raise arango.exceptions.IndexDeleteError: If delete fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/index/{}/{}'.format(self.name, index_id)
        )

        def response_handler(resp):
            if resp.error_code == 1212 and ignore_missing:
                return False
            if not resp.is_success:
                raise IndexDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def load_indexes(self):
        """Cache all indexes in the collection into memory.

        :return: True if index was loaded successfully.
        :rtype: bool
        :raise arango.exceptions.IndexLoadError: If operation fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/collection/{}/loadIndexesIntoMemory'.format(
                self.name
            )
        )

        def response_handler(resp):
            if not resp.is_success:
                raise IndexLoadError(resp, request)
            return True

        return await self._execute(request, response_handler)


class StandardCollection(Collection):
    """Standard ArangoDB collection API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    :param name: Collection name.
    :type name: str | unicode
    """

    def __init__(self, connection, executor, name):
        super(StandardCollection, self).__init__(connection, executor, name)

    def __repr__(self):
        return '<StandardCollection {}>'.format(self.name)

    async def get(self, document, rev=None, check_rev=True):
        """Return a document.

        :param document: Document ID, key or body. Document body must contain
            the "_id" or "_key" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :return: Document, or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, body, headers = self._prep_from_doc(document, rev, check_rev)

        request = Request(
            method='get',
            endpoint='/_api/document/{}'.format(handle),
            headers=headers,
            read=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202:
                return None
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def insert(
            self,
            document,
            return_new=False,
            sync=None,
            silent=False,
            overwrite=False,
            return_old=False,
            overwrite_mode=None,
            keep_none=None,
            merge=None):
        """Insert a new document.

        :param document: Document to insert. If it contains the "_key" or "_id"
            field, the value is used as the key of the new document (otherwise
            it is auto-generated). Any "_rev" field is ignored.
        :type document: dict
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param overwrite: If set to True, operation does not fail on duplicate
            key and existing document is overwritten (replace-insert).
        :type overwrite: bool
        :param return_old: Include body of the old document if overwritten.
            Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param overwrite_mode: Overwrite behavior used when the document key
            exists already. Allowed values are "replace" (replace-insert) or
            "update" (update-insert). Implicitly sets the value of parameter
            **overwrite**.
        :type overwrite_mode: str
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely. Applies
            only when **overwrite_mode** is set to "update" (update-insert).
        :type keep_none: bool
        :param merge: If set to True (default), sub-dictionaries are merged
            instead of the new one overwriting the old one. Applies only when
            **overwrite_mode** is set to "update" (update-insert).
        :type merge: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        document = self._ensure_key_from_id(document)

        params = {
            'returnNew': return_new,
            'silent': silent,
            'overwrite': overwrite,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync
        if overwrite_mode is not None:
            params['overwriteMode'] = overwrite_mode
        if keep_none is not None:
            params['keepNull'] = keep_none
        if merge is not None:
            params['mergeObjects'] = merge

        request = Request(
            method='post',
            endpoint='/_api/document/{}'.format(self.name),
            data=document,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentInsertError(resp, request)
            if silent:
                return True
            if '_oldRev' in resp.body:
                resp.body['_old_rev'] = resp.body.pop('_oldRev')
            return resp.body

        return await self._execute(request, response_handler)

    async def insert_many(
            self,
            documents,
            return_new=False,
            sync=None,
            silent=False,
            overwrite=False,
            return_old=False):
        """Insert multiple documents.

        .. note::

            If inserting a document fails, the exception is not raised but
            returned as an object in the result list. It is up to you to
            inspect the list to determine which documents were inserted
            successfully (returns document metadata) and which were not
            (returns exception object).

        :param documents: List of new documents to insert. If they contain the
            "_key" or "_id" fields, the values are used as the keys of the new
            documents (auto-generated otherwise). Any "_rev" field is ignored.
        :type documents: [dict]
        :param return_new: Include bodies of the new documents in the returned
            metadata. Ignored if parameter **silent** is set to True
        :type return_new: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param overwrite: If set to True, operation does not fail on duplicate
            keys and the existing documents are replaced.
        :type overwrite: bool
        :param return_old: Include body of the old documents if replaced.
            Applies only when value of **overwrite** is set to True.
        :type return_old: bool
        :return: List of document metadata (e.g. document keys, revisions) and
            any exception, or True if parameter **silent** was set to True.
        :rtype: [dict | ArangoError] | bool
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        documents = [self._ensure_key_from_id(doc) for doc in documents]

        params = {
            'returnNew': return_new,
            'silent': silent,
            'overwrite': overwrite,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='post',
            endpoint='/_api/document/{}'.format(self.name),
            data=documents,
            params=params
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentInsertError(resp, request)
            if silent is True:
                return True

            results = []
            for body in resp.body:
                if '_id' in body:
                    if '_oldRev' in body:
                        body['_old_rev'] = body.pop('_oldRev')
                    results.append(body)
                else:
                    sub_resp = self._conn.prep_bulk_err_response(resp, body)
                    results.append(DocumentInsertError(sub_resp, request))

            return results

        return await self._execute(request, response_handler)

    async def update(
            self,
            document,
            check_rev=True,
            merge=True,
            keep_none=True,
            return_new=False,
            return_old=False,
            sync=None,
            silent=False):
        """Update a document.

        :param document: Partial or full document with the updated values. It
            must contain the "_id" or "_key" field.
        :type document: dict
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param merge: If set to True, sub-dictionaries are merged instead of
            the new one overwriting the old one.
        :type merge: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely.
        :type keep_none: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        params = {
            'keepNull': keep_none,
            'mergeObjects': merge,
            'returnNew': return_new,
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='patch',
            endpoint='/_api/document/{}'.format(
                self._extract_id(document)
            ),
            data=document,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            elif not resp.is_success:
                raise DocumentUpdateError(resp, request)
            if silent is True:
                return True
            resp.body['_old_rev'] = resp.body.pop('_oldRev')
            return resp.body

        return await self._execute(request, response_handler)

    async def update_many(
            self,
            documents,
            check_rev=True,
            merge=True,
            keep_none=True,
            return_new=False,
            return_old=False,
            sync=None,
            silent=False):
        """Update multiple documents.

        .. note::

            If updating a document fails, the exception is not raised but
            returned as an object in the result list. It is up to you to
            inspect the list to determine which documents were updated
            successfully (returns document metadata) and which were not
            (returns exception object).

        :param documents: Partial or full documents with the updated values.
            They must contain the "_id" or "_key" fields.
        :type documents: [dict]
        :param check_rev: If set to True, revisions of **documents** (if given)
            are compared against the revisions of target documents.
        :type check_rev: bool
        :param merge: If set to True, sub-dictionaries are merged instead of
            the new ones overwriting the old ones.
        :type merge: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely.
        :type keep_none: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: List of document metadata (e.g. document keys, revisions) and
            any exceptions, or True if parameter **silent** was set to True.
        :rtype: [dict | ArangoError] | bool
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        """
        params = {
            'keepNull': keep_none,
            'mergeObjects': merge,
            'returnNew': return_new,
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        documents = [self._ensure_key_in_body(doc) for doc in documents]

        request = Request(
            method='patch',
            endpoint='/_api/document/{}'.format(self.name),
            data=documents,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentUpdateError(resp, request)
            if silent is True:
                return True

            results = []
            for body in resp.body:
                if '_id' in body:
                    body['_old_rev'] = body.pop('_oldRev')
                    results.append(body)
                else:
                    sub_resp = self._conn.prep_bulk_err_response(resp, body)
                    if sub_resp.error_code == 1200:
                        error = DocumentRevisionError(sub_resp, request)
                    else:  # pragma: no cover
                        error = DocumentUpdateError(sub_resp, request)
                    results.append(error)

            return results

        return await self._execute(request, response_handler)

    async def update_match(
            self,
            filters,
            body,
            limit=None,
            keep_none=True,
            sync=None,
            merge=True):
        """Update matching documents.

        :param filters: Document filters.
        :type filters: dict
        :param body: Full or partial document body with the updates.
        :type body: dict
        :param limit: Max number of documents to update. If the limit is lower
            than the number of matched documents, random documents are
            chosen. This parameter is not supported on sharded collections.
        :type limit: int
        :param keep_none: If set to True, fields with value None are retained
            in the document. Otherwise, they are removed completely.
        :type keep_none: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param merge: If set to True, sub-dictionaries are merged instead of
            the new ones overwriting the old ones.
        :type merge: bool
        :return: Number of documents updated.
        :rtype: int
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        """
        data = {
            'collection': self.name,
            'example': filters,
            'newValue': body,
            'keepNull': keep_none,
            'mergeObjects': merge
        }
        if limit is not None:
            data['limit'] = limit
        if sync is not None:
            data['waitForSync'] = sync

        request = Request(
            method='put',
            endpoint='/_api/simple/update-by-example',
            data=data,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentUpdateError(resp, request)
            return resp.body['updated']

        return await self._execute(request, response_handler)

    async def replace(
            self,
            document,
            check_rev=True,
            return_new=False,
            return_old=False,
            sync=None,
            silent=False):
        """Replace a document.

        :param document: New document to replace the old one with. It must
            contain the "_id" or "_key" field. Edge document must also have
            "_from" and "_to" fields.
        :type document: dict
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        params = {
            'returnNew': return_new,
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='put',
            endpoint='/_api/document/{}'.format(
                self._extract_id(document)
            ),
            params=params,
            data=document,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentReplaceError(resp, request)
            if silent is True:
                return True
            resp.body['_old_rev'] = resp.body.pop('_oldRev')
            return resp.body

        return await self._execute(request, response_handler)

    async def replace_many(
            self,
            documents,
            check_rev=True,
            return_new=False,
            return_old=False,
            sync=None,
            silent=False):
        """Replace multiple documents.

        .. note::

            If replacing a document fails, the exception is not raised but
            returned as an object in the result list. It is up to you to
            inspect the list to determine which documents were replaced
            successfully (returns document metadata) and which were not
            (returns exception object).

        :param documents: New documents to replace the old ones with. They must
            contain the "_id" or "_key" fields. Edge documents must also have
            "_from" and "_to" fields.
        :type documents: [dict]
        :param check_rev: If set to True, revisions of **documents** (if given)
            are compared against the revisions of target documents.
        :type check_rev: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: List of document metadata (e.g. document keys, revisions) and
            any exceptions, or True if parameter **silent** was set to True.
        :rtype: [dict | ArangoError] | bool
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        """
        params = {
            'returnNew': return_new,
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        documents = [self._ensure_key_in_body(doc) for doc in documents]

        request = Request(
            method='put',
            endpoint='/_api/document/{}'.format(self.name),
            params=params,
            data=documents,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentReplaceError(resp, request)
            if silent is True:
                return True

            results = []
            for body in resp.body:
                if '_id' in body:
                    body['_old_rev'] = body.pop('_oldRev')
                    results.append(body)
                else:
                    sub_resp = self._conn.prep_bulk_err_response(resp, body)
                    if sub_resp.error_code == 1200:
                        error = DocumentRevisionError(sub_resp, request)
                    else:  # pragma: no cover
                        error = DocumentReplaceError(sub_resp, request)
                    results.append(error)

            return results

        return await self._execute(request, response_handler)

    async def replace_match(self, filters, body, limit=None, sync=None):
        """Replace matching documents.

        :param filters: Document filters.
        :type filters: dict
        :param body: New document body.
        :type body: dict
        :param limit: Max number of documents to replace. If the limit is lower
            than the number of matched documents, random documents are chosen.
        :type limit: int
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :return: Number of documents replaced.
        :rtype: int
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        """
        data = {
            'collection': self.name,
            'example': filters,
            'newValue': body
        }
        if limit is not None:
            data['limit'] = limit
        if sync is not None:
            data['waitForSync'] = sync

        request = Request(
            method='put',
            endpoint='/_api/simple/replace-by-example',
            data=data,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentReplaceError(resp, request)
            return resp.body['replaced']

        return await self._execute(request, response_handler)

    async def delete(
            self,
            document,
            rev=None,
            check_rev=True,
            ignore_missing=False,
            return_old=False,
            sync=None,
            silent=False):
        """Delete a document.

        :param document: Document ID, key or body. Document body must contain
            the "_id" or "_key" field.
        :type document: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **document** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **document** (if given)
            is compared against the revision of target document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision), or True if
            parameter **silent** was set to True, or False if document was not
            found and **ignore_missing** was set to True (does not apply in
            transactions).
        :rtype: bool | dict
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, body, headers = self._prep_from_doc(document, rev, check_rev)

        params = {
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='delete',
            endpoint='/_api/document/{}'.format(handle),
            params=params,
            headers=headers,
            write=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202 and ignore_missing:
                return False
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentDeleteError(resp, request)
            return True if silent else resp.body

        return await self._execute(request, response_handler)

    async def delete_many(
            self,
            documents,
            return_old=False,
            check_rev=True,
            sync=None,
            silent=False):
        """Delete multiple documents.

        .. note::

            If deleting a document fails, the exception is not raised but
            returned as an object in the result list. It is up to you to
            inspect the list to determine which documents were deleted
            successfully (returns document metadata) and which were not
            (returns exception object).

        :param documents: Document IDs, keys or bodies. Document bodies must
            contain the "_id" or "_key" fields.
        :type documents: [str | unicode | dict]
        :param return_old: Include bodies of the old documents in the result.
        :type return_old: bool
        :param check_rev: If set to True, revisions of **documents** (if given)
            are compared against the revisions of target documents.
        :type check_rev: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: List of document metadata (e.g. document keys, revisions) and
            any exceptions, or True if parameter **silent** was set to True.
        :rtype: [dict | ArangoError] | bool
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        """
        params = {
            'returnOld': return_old,
            'ignoreRevs': not check_rev,
            'overwrite': not check_rev,
            'silent': silent
        }
        if sync is not None:
            params['waitForSync'] = sync

        documents = [
            self._ensure_key_in_body(doc) if isinstance(doc, dict) else doc
            for doc in documents
        ]

        request = Request(
            method='delete',
            endpoint='/_api/document/{}'.format(self.name),
            params=params,
            data=documents,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentDeleteError(resp, request)
            if silent is True:
                return True

            results = []
            for body in resp.body:
                if '_id' in body:
                    results.append(body)
                else:
                    sub_resp = self._conn.prep_bulk_err_response(resp, body)
                    if sub_resp.error_code == 1200:
                        error = DocumentRevisionError(sub_resp, request)
                    else:
                        error = DocumentDeleteError(sub_resp, request)
                    results.append(error)

            return results

        return await self._execute(request, response_handler)

    async def delete_match(self, filters, limit=None, sync=None):
        """Delete matching documents.

        :param filters: Document filters.
        :type filters: dict
        :param limit: Max number of documents to delete. If the limit is lower
            than the number of matched documents, random documents are chosen.
        :type limit: int
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :return: Number of documents deleted.
        :rtype: dict
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        """
        data = {'collection': self.name, 'example': filters}
        if sync is not None:
            data['waitForSync'] = sync
        if limit is not None and limit != 0:
            data['limit'] = limit

        request = Request(
            method='put',
            endpoint='/_api/simple/remove-by-example',
            data=data,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentDeleteError(resp, request)
            return resp.body['deleted']

        return await self._execute(request, response_handler)

    async def import_bulk(
            self,
            documents,
            halt_on_error=True,
            details=True,
            from_prefix=None,
            to_prefix=None,
            overwrite=None,
            on_duplicate=None,
            sync=None):
        """Insert multiple documents into the collection.

        This method is faster than
        :func:`arango.collection.StandardCollection.insert_many` but does not
        return as many details.

        :param documents: List of new documents to insert. If they contain the
            "_key" or "_id" fields, the values are used as the keys of the new
            documents (auto-generated otherwise). Any "_rev" field is ignored.
        :type documents: [dict]
        :param halt_on_error: Halt the entire import on an error.
        :type halt_on_error: bool
        :param details: If set to True, the returned result will include an
            additional list of detailed error messages.
        :type details: bool
        :param from_prefix: String prefix prepended to the value of "_from"
            field in each edge document inserted. For example, prefix "foo"
            prepended to "_from": "bar" will result in "_from": "foo/bar".
            Applies only to edge collections.
        :type from_prefix: str | unicode
        :param to_prefix: String prefix prepended to the value of "_to" field
            in edge document inserted. For example, prefix "foo" prepended to
            "_to": "bar" will result in "_to": "foo/bar". Applies only to edge
            collections.
        :type to_prefix: str | unicode
        :param overwrite: If set to True, all existing documents are removed
            prior to the import. Indexes are still preserved.
        :type overwrite: bool
        :param on_duplicate: Action to take on unique key constraint violations
            (for documents with "_key" fields). Allowed values are "error" (do
            not import the new documents and count them as errors), "update"
            (update the existing documents while preserving any fields missing
            in the new ones), "replace" (replace the existing documents with
            new ones), and  "ignore" (do not import the new documents and count
            them as ignored, as opposed to counting them as errors). Options
            "update" and "replace" may fail on secondary unique key constraint
            violations.
        :type on_duplicate: str | unicode
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :return: Result of the bulk import.
        :rtype: dict
        :raise arango.exceptions.DocumentInsertError: If import fails.
        """
        documents = [self._ensure_key_from_id(doc) for doc in documents]

        params = {
            'type': 'array',
            'collection': self.name
        }
        if halt_on_error is not None:
            params['complete'] = halt_on_error
        if details is not None:
            params['details'] = details
        if from_prefix is not None:  # pragma: no cover
            params['fromPrefix'] = from_prefix
        if to_prefix is not None:  # pragma: no cover
            params['toPrefix'] = to_prefix
        if overwrite is not None:
            params['overwrite'] = overwrite
        if on_duplicate is not None:
            params['onDuplicate'] = on_duplicate
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='post',
            endpoint='/_api/import',
            data=documents,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentInsertError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)


class VertexCollection(Collection):
    """Vertex collection API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    :param graph: Graph name.
    :type graph: str | unicode
    :param name: Vertex collection name.
    :type name: str | unicode
    """

    def __init__(self, connection, executor, graph, name):
        super(VertexCollection, self).__init__(connection, executor, name)
        self._graph = graph

    def __repr__(self):
        return '<VertexCollection {}>'.format(self.name)

    @property
    def graph(self):
        """Return the graph name.

        :return: Graph name.
        :rtype: str | unicode
        """
        return self._graph

    async def get(self, vertex, rev=None, check_rev=True):
        """Return a vertex document.

        :param vertex: Vertex document ID, key or body. Document body must
            contain the "_id" or "_key" field.
        :type vertex: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **vertex** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :return: Vertex document or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, body, headers = self._prep_from_doc(vertex, rev, check_rev)

        request = Request(
            method='get',
            endpoint='/_api/gharial/{}/vertex/{}'.format(
                self._graph, handle
            ),
            headers=headers,
            read=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202:
                return None
            if resp.status_code == 412:
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return resp.body['vertex']

        return await self._execute(request, response_handler)

    async def insert(self, vertex, sync=None, silent=False, return_new=False):
        """Insert a new vertex document.

        :param vertex: New vertex document to insert. If it has "_key" or "_id"
            field, its value is used as key of the new vertex (otherwise it is
            auto-generated). Any "_rev" field is ignored.
        :type vertex: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision), or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        vertex = self._ensure_key_from_id(vertex)

        params = {
            'silent': silent,
            'returnNew': return_new
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='post',
            endpoint='/_api/gharial/{}/vertex/{}'.format(
                self._graph, self.name
            ),
            data=vertex,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentInsertError(resp, request)
            if silent:
                return True
            return format_vertex(resp.body)

        return await self._execute(request, response_handler)

    async def update(self,
               vertex,
               check_rev=True,
               keep_none=True,
               sync=None,
               silent=False,
               return_old=False,
               return_new=False):
        """Update a vertex document.

        :param vertex: Partial or full vertex document with updated values. It
            must contain the "_key" or "_id" field.
        :type vertex: dict
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. If set to False, they are removed completely.
        :type keep_none: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        vertex_id, headers = self._prep_from_body(vertex, check_rev)

        params = {
            'keepNull': keep_none,
            'overwrite': not check_rev,
            'silent': silent,
            'returnNew': return_new,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='patch',
            endpoint='/_api/gharial/{}/vertex/{}'.format(
                self._graph, vertex_id
            ),
            headers=headers,
            params=params,
            data=vertex,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            elif not resp.is_success:
                raise DocumentUpdateError(resp, request)
            if silent is True:
                return True
            return format_vertex(resp.body)

        return await self._execute(request, response_handler)

    async def replace(self,
                vertex,
                check_rev=True,
                sync=None,
                silent=False,
                return_old=False,
                return_new=False):
        """Replace a vertex document.

        :param vertex: New vertex document to replace the old one with. It must
            contain the "_key" or "_id" field.
        :type vertex: dict
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        vertex_id, headers = self._prep_from_body(vertex, check_rev)

        params = {
            'silent': silent,
            'returnNew': return_new,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='put',
            endpoint='/_api/gharial/{}/vertex/{}'.format(
                self._graph, vertex_id
            ),
            headers=headers,
            params=params,
            data=vertex,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            elif not resp.is_success:
                raise DocumentReplaceError(resp, request)
            if silent is True:
                return True
            return format_vertex(resp.body)

        return await self._execute(request, response_handler)

    async def delete(self,
               vertex,
               rev=None,
               check_rev=True,
               ignore_missing=False,
               sync=None,
               return_old=False):
        """Delete a vertex document.

        :param vertex: Vertex document ID, key or body. Document body must
            contain the "_id" or "_key" field.
        :type vertex: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **vertex** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param return_old: Return body of the old document in the result.
        :type return_old: bool
        :return: True if vertex was deleted successfully, False if vertex was
            not found and **ignore_missing** was set to True (does not apply in
            transactions). Old document is returned if **return_old** is set to
            True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, _, headers = self._prep_from_doc(vertex, rev, check_rev)

        params = {'returnOld': return_old}
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='delete',
            endpoint='/_api/gharial/{}/vertex/{}'.format(
                self._graph, handle
            ),
            params=params,
            headers=headers,
            write=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202 and ignore_missing:
                return False
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentDeleteError(resp, request)
            if 'old' in resp.body:
                return {'old': resp.body['old']}
            return True

        return await self._execute(request, response_handler)


class EdgeCollection(Collection):
    """ArangoDB edge collection API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    :param graph: Graph name.
    :type graph: str | unicode
    :param name: Edge collection name.
    :type name: str | unicode
    """

    def __init__(self, connection, executor, graph, name):
        super(EdgeCollection, self).__init__(connection, executor, name)
        self._graph = graph

    def __repr__(self):
        return '<EdgeCollection {}>'.format(self.name)

    @property
    def graph(self):
        """Return the graph name.

        :return: Graph name.
        :rtype: str | unicode
        """
        return self._graph

    async def get(self, edge, rev=None, check_rev=True):
        """Return an edge document.

        :param edge: Edge document ID, key or body. Document body must contain
            the "_id" or "_key" field.
        :type edge: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **edge** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :return: Edge document or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, body, headers = self._prep_from_doc(edge, rev, check_rev)

        request = Request(
            method='get',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._graph, handle
            ),
            headers=headers,
            read=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202:
                return None
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentGetError(resp, request)
            return resp.body['edge']

        return await self._execute(request, response_handler)

    async def insert(self, edge, sync=None, silent=False, return_new=False):
        """Insert a new edge document.

        :param edge: New edge document to insert. It must contain "_from" and
            "_to" fields. If it has "_key" or "_id" field, its value is used
            as key of the new edge document (otherwise it is auto-generated).
            Any "_rev" field is ignored.
        :type edge: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        edge = self._ensure_key_from_id(edge)

        params = {
            'silent': silent,
            'returnNew': return_new
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='post',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._graph, self.name
            ),
            data=edge,
            params=params,
            write=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise DocumentInsertError(resp, request)
            if silent:
                return True
            return format_edge(resp.body)

        return await self._execute(request, response_handler)

    async def update(
            self,
            edge,
            check_rev=True,
            keep_none=True,
            sync=None,
            silent=False,
            return_old=False,
            return_new=False):
        """Update an edge document.

        :param edge: Partial or full edge document with updated values. It must
            contain the "_key" or "_id" field.
        :type edge: dict
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. If set to False, they are removed completely.
        :type keep_none: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        edge_id, headers = self._prep_from_body(edge, check_rev)

        params = {
            'keepNull': keep_none,
            'overwrite': not check_rev,
            'silent': silent,
            'returnNew': return_new,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='patch',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._graph, edge_id
            ),
            headers=headers,
            params=params,
            data=edge,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentUpdateError(resp, request)
            if silent is True:
                return True
            return format_edge(resp.body)

        return await self._execute(request, response_handler)

    async def replace(
            self,
            edge,
            check_rev=True,
            sync=None,
            silent=False,
            return_old=False,
            return_new=False):
        """Replace an edge document.

        :param edge: New edge document to replace the old one with. It must
            contain the "_key" or "_id" field. It must also contain the "_from"
            and "_to" fields.
        :type edge: dict
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_old: Include body of the old document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_old: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        edge_id, headers = self._prep_from_body(edge, check_rev)

        params = {
            'silent': silent,
            'returnNew': return_new,
            'returnOld': return_old
        }
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='put',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._graph, edge_id
            ),
            headers=headers,
            params=params,
            data=edge,
            write=self.name
        )

        def response_handler(resp):
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentReplaceError(resp, request)
            if silent is True:
                return True
            return format_edge(resp.body)

        return await self._execute(request, response_handler)

    async def delete(
            self,
            edge,
            rev=None,
            check_rev=True,
            ignore_missing=False,
            sync=None,
            return_old=False):
        """Delete an edge document.

        :param edge: Edge document ID, key or body. Document body must contain
            the "_id" or "_key" field.
        :type edge: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **edge** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param return_old: Return body of the old document in the result.
        :type return_old: bool
        :return: True if edge was deleted successfully, False if edge was not
            found and **ignore_missing** was set to True (does not  apply in
            transactions).
        :rtype: bool
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        handle, _, headers = self._prep_from_doc(edge, rev, check_rev)

        params = {'returnOld': return_old}
        if sync is not None:
            params['waitForSync'] = sync

        request = Request(
            method='delete',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._graph, handle
            ),
            params=params,
            headers=headers,
            write=self.name
        )

        def response_handler(resp):
            if resp.error_code == 1202 and ignore_missing:
                return False
            if resp.status_code == 412:  # pragma: no cover
                raise DocumentRevisionError(resp, request)
            if not resp.is_success:
                raise DocumentDeleteError(resp, request)
            if 'old' in resp.body:
                return {'old': resp.body['old']}
            return True

        return await self._execute(request, response_handler)

    async def link(self,
             from_vertex,
             to_vertex,
             data=None,
             sync=None,
             silent=False,
             return_new=False):
        """Insert a new edge document linking the given vertices.

        :param from_vertex: "From" vertex document ID or body with "_id" field.
        :type from_vertex: str | unicode | dict
        :param to_vertex: "To" vertex document ID or body with "_id" field.
        :type to_vertex: str | unicode | dict
        :param data: Any extra data for the new edge document. If it has "_key"
            or "_id" field, its value is used as key of the new edge document
            (otherwise it is auto-generated).
        :type data: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :param return_new: Include body of the new document in the returned
            metadata. Ignored if parameter **silent** is set to True.
        :type return_new: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        edge = {
            '_from': get_doc_id(from_vertex),
            '_to': get_doc_id(to_vertex)
        }
        if data is not None:
            edge.update(self._ensure_key_from_id(data))
        return await self.insert(
            edge,
            sync=sync,
            silent=silent,
            return_new=return_new
        )

    async def edges(self, vertex, direction=None):
        """Return the edge documents coming in and/or out of the vertex.

        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :param direction: The direction of the edges. Allowed values are "in"
            and "out". If not set, edges in both directions are returned.
        :type direction: str | unicode
        :return: List of edges and statistics.
        :rtype: dict
        :raise arango.exceptions.EdgeListError: If retrieval fails.
        """
        params = {'vertex': get_doc_id(vertex)}
        if direction is not None:
            params['direction'] = direction

        request = Request(
            method='get',
            endpoint='/_api/edges/{}'.format(self.name),
            params=params,
            read=self.name
        )

        def response_handler(resp):
            if not resp.is_success:
                raise EdgeListError(resp, request)
            stats = resp.body['stats']
            return {
                'edges': resp.body['edges'],
                'stats': {
                    'filtered': stats['filtered'],
                    'scanned_index': stats['scannedIndex'],
                }
            }

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

import pytest
from six import string_types

from aioarangodb.exceptions import (
    DatabasePropertiesError,
    UserCreateError,
    UserDeleteError,
    UserGetError,
    UserListError,
    UserReplaceError,
    UserUpdateError,
)
from aioarangodb.tests.helpers import (
    assert_raises,
    extract,
    generate_db_name,
    generate_username,
    generate_string,
)
pytestmark = pytest.mark.asyncio


async def test_user_management(sys_db, bad_db):
    # Test create user
    username = generate_username()
    password = generate_string()
    assert not await sys_db.has_user(username)

    new_user = await sys_db.create_user(
        username=username,
        password=password,
        active=True,
        extra={'foo': 'bar'},
    )
    assert new_user['username'] == username
    assert new_user['active'] is True
    assert new_user['extra'] == {'foo': 'bar'}
    assert await sys_db.has_user(username)

    # Test create duplicate user
    with assert_raises(UserCreateError) as err:
        await sys_db.create_user(
            username=username,
            password=password
        )
    assert err.value.error_code == 1702

    # Test list users
    for user in await sys_db.users():
        assert isinstance(user['username'], string_types)
        assert isinstance(user['active'], bool)
        assert isinstance(user['extra'], dict)
    assert await sys_db.user(username) == new_user

    # Test list users with bad database
    with assert_raises(UserListError) as err:
        await bad_db.users()
    assert err.value.error_code in {11, 1228}

    # Test get user
    users = await sys_db.users()
    for user in users:
        assert 'active' in user
        assert 'extra' in user
        assert 'username' in user
    assert username in await extract('username', await sys_db.users())

    # Test get missing user
    with assert_raises(UserGetError) as err:
        await sys_db.user(generate_username())
    assert err.value.error_code == 1703

    # Update existing user
    new_user = await sys_db.update_user(
        username=username,
        password=password,
        active=False,
        extra={'bar': 'baz'},
    )
    assert new_user['username'] == username
    assert new_user['active'] is False
    assert new_user['extra'] == {'bar': 'baz'}
    assert await sys_db.user(username) == new_user

    # Update missing user
    with assert_raises(UserUpdateError) as err:
        await sys_db.update_user(
            username=generate_username(),
            password=generate_string()
        )
    assert err.value.error_code == 1703

    # Replace existing user
    new_user = await sys_db.replace_user(
        username=username,
        password=password,
        active=False,
        extra={'baz': 'qux'},
    )
    assert new_user['username'] == username
    assert new_user['active'] is False
    assert new_user['extra'] == {'baz': 'qux'}
    assert await sys_db.user(username) == new_user

    # Replace missing user
    with assert_raises(UserReplaceError) as err:
        await sys_db.replace_user(
            username=generate_username(),
            password=generate_string()
        )
    assert err.value.error_code == 1703

    # Delete an existing user
    assert await sys_db.delete_user(username) is True

    # Delete a missing user
    with assert_raises(UserDeleteError) as err:
        await sys_db.delete_user(username, ignore_missing=False)
    assert err.value.error_code == 1703
    assert await sys_db.delete_user(username, ignore_missing=True) is False


async def test_user_change_password(client, sys_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    username = generate_username()
    password1 = generate_string()
    password2 = generate_string()

    await sys_db.create_user(username, password1)
    await sys_db.update_permission(username, 'rw', sys_db.name)

    db1 = await client.db(sys_db.name, username, password1)
    db2 = await client.db(sys_db.name, username, password2)

    # Check authentication
    assert isinstance(await db1.properties(), dict)
    with assert_raises(DatabasePropertiesError) as err:
        await db2.properties()
    assert err.value.http_code == 401

    # Update the user password and check again
    await sys_db.update_user(username, password2)
    assert isinstance(await db2.properties(), dict)
    with assert_raises(DatabasePropertiesError) as err:
        await db1.properties()
    assert err.value.http_code == 401

    # Replace the user password back and check again
    await sys_db.update_user(username, password1)
    assert isinstance(await db1.properties(), dict)
    with assert_raises(DatabasePropertiesError) as err:
        await db2.properties()
    assert err.value.http_code == 401


async def test_user_create_with_new_database(client, sys_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    db_name = generate_db_name()

    username1 = generate_username()
    username2 = generate_username()
    username3 = generate_username()

    password1 = generate_string()
    password2 = generate_string()
    password3 = generate_string()

    result = await sys_db.create_database(
        name=db_name,
        users=[
            {'username': username1, 'password': password1, 'active': True},
            {'username': username2, 'password': password2, 'active': True},
            {'username': username3, 'password': password3, 'active': False},
        ]
    )
    assert result is True

    await sys_db.update_permission(username1, permission='rw', database=db_name)
    await sys_db.update_permission(username2, permission='rw', database=db_name)
    await sys_db.update_permission(username3, permission='rw', database=db_name)

    # Test if the users were created properly
    usernames = await extract('username', await sys_db.users())
    assert all(u in usernames for u in [username1, username2, username3])

    # Test if the first user has access to the database
    db = await client.db(db_name, username1, password1)
    await db.properties()

    # Test if the second user also has access to the database
    db = await client.db(db_name, username2, password2)
    await db.properties()

    # Test if the third user has access to the database (should not)
    db = await client.db(db_name, username3, password3)
    with assert_raises(DatabasePropertiesError) as err:
        await db.properties()
    assert err.value.http_code == 401



from __future__ import absolute_import, unicode_literals
import pytest

from aioarangodb.exceptions import (
    AnalyzerCreateError,
    AnalyzerDeleteError,
    AnalyzerGetError,
    AnalyzerListError
)
from aioarangodb.tests.helpers import assert_raises, generate_analyzer_name


@pytest.mark.asyncio
async def test_analyzer_management(db, bad_db, cluster):
    analyzer_name = generate_analyzer_name()
    full_analyzer_name = db.name + '::' + analyzer_name
    bad_analyzer_name = generate_analyzer_name()

    # Test create analyzer
    result = await db.create_analyzer(analyzer_name, 'identity', {})
    assert result['name'] == full_analyzer_name
    assert result['type'] == 'identity'
    assert result['properties'] == {}
    assert result['features'] == []

    # Test create duplicate with bad database
    with assert_raises(AnalyzerCreateError) as err:
        await bad_db.create_analyzer(analyzer_name, 'identity', {}, [])
    assert err.value.error_code in {11, 1228}

    # Test get analyzer
    result = await db.analyzer(analyzer_name)
    assert result['name'] == full_analyzer_name
    assert result['type'] == 'identity'
    assert result['properties'] == {}
    assert result['features'] == []

    # Test get missing analyzer
    with assert_raises(AnalyzerGetError) as err:
        await db.analyzer(bad_analyzer_name)
    assert err.value.error_code in {1202}

    # Test list analyzers
    result = await db.analyzers()
    assert full_analyzer_name in [a['name'] for a in result]

    # Test list analyzers with bad database
    with assert_raises(AnalyzerListError) as err:
        await bad_db.analyzers()
    assert err.value.error_code in {11, 1228}

    # Test delete analyzer
    assert await db.delete_analyzer(analyzer_name, force=True) is True
    assert full_analyzer_name not in [a['name'] for a in await db.analyzers()]

    # Test delete missing analyzer
    with assert_raises(AnalyzerDeleteError) as err:
        await db.delete_analyzer(analyzer_name)
    assert err.value.error_code in {1202}

    # Test delete missing analyzer with ignore_missing set to True
    assert await db.delete_analyzer(analyzer_name, ignore_missing=True) is False



from __future__ import absolute_import, unicode_literals

__all__ = ['Cursor']

from collections import deque

from .exceptions import (
    CursorCloseError,
    CursorEmptyError,
    CursorNextError,
    CursorStateError,
    CursorCountError
)
from .request import Request


class Cursor(object):
    """Cursor API wrapper.

    Cursors fetch query results from ArangoDB server in batches. Cursor objects
    are *stateful* as they store the fetched items in-memory. They must not be
    shared across threads without proper locking mechanism.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param init_data: Cursor initialization data.
    :type init_data: dict | list
    :param cursor_type: Cursor type ("cursor" or "export").
    :type cursor_type: str | unicode
    """

    __slots__ = [
        '_conn',
        '_type',
        '_id',
        '_count',
        '_cached',
        '_stats',
        '_profile',
        '_warnings',
        '_has_more',
        '_batch'
    ]

    def __init__(self, connection, init_data, cursor_type='cursor'):
        self._conn = connection
        self._type = cursor_type
        self._batch = deque()
        self._id = None
        self._count = None
        self._cached = None
        self._stats = None
        self._profile = None
        self._warnings = None
        self._update(init_data)

    def __aiter__(self):
        return self

    async def __anext__(self):  # pragma: no cover
        return await self.next()

    async def __aenter__(self):
        return self

    def __len__(self):
        if self._count is None:
            raise CursorCountError('cursor count not enabled')
        return self._count

    async def __aexit__(self, *_):
        await self.close(ignore_missing=True)

    def __repr__(self):
        return '<Cursor {}>'.format(self._id) if self._id else '<Cursor>'

    def _update(self, data):
        """Update the cursor using data from ArangoDB server.

        :param data: Cursor data from ArangoDB server (e.g. results).
        :type data: dict
        """
        result = {}

        if 'id' in data:
            self._id = data['id']
            result['id'] = data['id']
        if 'count' in data:
            self._count = data['count']
            result['count'] = data['count']
        if 'cached' in data:
            self._cached = data['cached']
            result['cached'] = data['cached']

        self._has_more = data['hasMore']
        result['has_more'] = data['hasMore']

        self._batch.extend(data['result'])
        result['batch'] = data['result']

        if 'extra' in data:
            extra = data['extra']

            if 'profile' in extra:
                self._profile = extra['profile']
                result['profile'] = extra['profile']

            if 'warnings' in extra:
                self._warnings = extra['warnings']
                result['warnings'] = extra['warnings']

            if 'stats' in extra:
                stats = extra['stats']
                if 'writesExecuted' in stats:
                    stats['modified'] = stats.pop('writesExecuted')
                if 'writesIgnored' in stats:
                    stats['ignored'] = stats.pop('writesIgnored')
                if 'scannedFull' in stats:
                    stats['scanned_full'] = stats.pop('scannedFull')
                if 'scannedIndex' in stats:
                    stats['scanned_index'] = stats.pop('scannedIndex')
                if 'executionTime' in stats:
                    stats['execution_time'] = stats.pop('executionTime')
                if 'httpRequests' in stats:
                    stats['http_requests'] = stats.pop('httpRequests')
                self._stats = stats
                result['statistics'] = stats

        return result

    @property
    def id(self):
        """Return the cursor ID.

        :return: Cursor ID.
        :rtype: str | unicode
        """
        return self._id

    @property
    def type(self):
        """Return the cursor type.

        :return: Cursor type ("cursor" or "export").
        :rtype: str | unicode
        """
        return self._type

    def batch(self):
        """Return the current batch of results.

        :return: Current batch.
        :rtype: collections.deque
        """
        return self._batch

    def has_more(self):
        """Return True if more results are available on the server.

        :return: True if more results are available on the server.
        :rtype: bool
        """
        return self._has_more

    def count(self):
        """Return the total number of documents in the entire result set.

        :return: Total number of documents, or None if the count option
            was not enabled during cursor initialization.
        :rtype: int | None
        """
        return self._count

    def cached(self):
        """Return True if results are cached.

        :return: True if results are cached.
        :rtype: bool
        """
        return self._cached

    def statistics(self):
        """Return cursor statistics.

        :return: Cursor statistics.
        :rtype: dict
        """
        return self._stats

    def profile(self):
        """Return cursor performance profile.

        :return: Cursor performance profile.
        :rtype: dict
        """
        return self._profile

    def warnings(self):
        """Return any warnings from the query execution.

        :return: Warnings, or None if there are none.
        :rtype: list
        """
        return self._warnings

    def empty(self):
        """Check if the current batch is empty.

        :return: True if current batch is empty, False otherwise.
        :rtype: bool
        """
        return len(self._batch) == 0

    async def next(self):
        """Pop the next item from the current batch.

        If current batch is empty/depleted, an API request is automatically
        sent to ArangoDB server to fetch the next batch and update the cursor.

        :return: Next item in current batch.
        :rtype: str | unicode | bool | int | list | dict
        :raise StopAsyncIteration: If the result set is depleted.
        :raise arango.exceptions.CursorNextError: If batch retrieval fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self.empty():
            if not self.has_more():
                raise StopAsyncIteration
            await self.fetch()

        return self.pop()

    def pop(self):
        """Pop the next item from current batch.

        If current batch is empty/depleted, an exception is raised. You must
        call :func:`arango.cursor.Cursor.fetch` to manually fetch the next
        batch from server.

        :return: Next item in current batch.
        :rtype: str | unicode | bool | int | list | dict
        :raise arango.exceptions.CursorEmptyError: If current batch is empty.
        """
        if len(self._batch) == 0:
            raise CursorEmptyError('current batch is empty')
        return self._batch.popleft()

    async def fetch(self):
        """Fetch the next batch from server and update the cursor.

        :return: New batch details.
        :rtype: dict
        :raise arango.exceptions.CursorNextError: If batch retrieval fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self._id is None:
            raise CursorStateError('cursor ID not set')
        request = Request(
            method='put',
            endpoint='/_api/{}/{}'.format(self._type, self._id)
        )
        resp = await self._conn.send_request(request)

        if not resp.is_success:
            raise CursorNextError(resp, request)
        return self._update(resp.body)

    async def close(self, ignore_missing=False):
        """Close the cursor and free any server resources tied to it.

        :param ignore_missing: Do not raise exception on missing cursors.
        :type ignore_missing: bool
        :return: True if cursor was closed successfully, False if cursor was
            missing on the server and **ignore_missing** was set to True, None
            if there are no cursors to close server-side (e.g. result set is
            smaller than the batch size).
        :rtype: bool | None
        :raise arango.exceptions.CursorCloseError: If operation fails.
        :raise arango.exceptions.CursorStateError: If cursor ID is not set.
        """
        if self._id is None:
            return None
        request = Request(
            method='delete',
            endpoint='/_api/{}/{}'.format(self._type, self._id)
        )
        resp = await self._conn.send_request(request)
        if resp.is_success:
            return True
        if resp.status_code == 404 and ignore_missing:
            return False
        raise CursorCloseError(resp, request)



##################
# General Errors #
##################

# No error occurred.
NO_ERROR = 0

# General error occurred.
FAILED = 1

# Operating system error occurred.
SYS_ERROR = 2

# Out of memory.
OUT_OF_MEMORY = 3

# Internal error occurred.
INTERNAL = 4

# Illegal number representation given.
ILLEGAL_NUMBER = 5

# Numeric overflow occurred.
NUMERIC_OVERFLOW = 6

# Unknown option supplied by user.
ILLEGAL_OPTION = 7

# Detected PID without living process.
DEAD_PID = 8

# Feature not implemented.
NOT_IMPLEMENTED = 9

# Bad parameter.
BAD_PARAMETER = 10

# Missing permission.
FORBIDDEN = 11

# Out of memory (mmap).
OUT_OF_MEMORY_MMAP = 12

# Corrupt CSV line.
CORRUPTED_CSV = 13

# File not found.
FILE_NOT_FOUND = 14

# Cannot write to file.
CANNOT_WRITE_FILE = 15

# Cannot overwrite file.
CANNOT_OVERWRITE_FILE = 16

# Type error occurred.
TYPE_ERROR = 17

# Timed out waiting for a lock.
LOCK_TIMEOUT = 18

# Cannot create a directory.
CANNOT_CREATE_DIRECTORY = 19

# Cannot create a temporary file.
CANNOT_CREATE_TEMP_FILE = 20

# Request cancelled by user.
REQUEST_CANCELED = 21

# Raised for debugging.
DEBUG = 22

# Invalid IP address.
IP_ADDRESS_INVALID = 25

# File exists already.
FILE_EXISTS = 27

# Locked resource or operation.
LOCKED = 28

# Deadlock detected when accessing collections.
DEADLOCK = 29

# Call failed as server shutdown is in progress.
SHUTTING_DOWN = 30

# Feature only for enterprise version of ArangoDB.
ONLY_ENTERPRISE = 31

# Resource usage exceeded maximum value.
RESOURCE_LIMIT = 32

# ICU operation failed.
ICU_ERROR = 33

# Cannot read a file.
CANNOT_READ_FILE = 34

# Incompatible version of ArangoDB.
INCOMPATIBLE_VERSION = 35

# Requested resource disabled.
DISABLED = 36

###########################
# HTTP Error Status Codes #
###########################

# Bad HTTP parameter.
HTTP_BAD_PARAMETER = 400

# User unauthorized.
HTTP_UNAUTHORIZED = 401

# Operation forbidden.
HTTP_FORBIDDEN = 403

# Unknown URI.
HTTP_NOT_FOUND = 404

# HTTP method unknown.
HTTP_METHOD_NOT_ALLOWED = 405

# HTTP content type not supported.
HTTP_NOT_ACCEPTABLE = 406

# Precondition not met.
HTTP_PRECONDITION_FAILED = 412

# Internal server error occurred.
HTTP_SERVER_ERROR = 500

# Service temporarily unavailable.
HTTP_SERVICE_UNAVAILABLE = 503

# Service contacted by ArangoDB did not respond in time.
HTTP_GATEWAY_TIMEOUT = 504

##########################
# HTTP Processing Errors #
##########################

# Corrupted JSON string.
HTTP_CORRUPTED_JSON = 600

# URL contains superfluous suffices.
HTTP_SUPERFLUOUS_SUFFICES = 601

####################################
# Internal ArangoDB Storage Errors #
####################################

# Datafile in illegal state.
ILLEGAL_STATE = 1000

# User attempted to write to a sealed datafile.
DATAFILE_SEALED = 1002

# Read-only datafile or collection.
READ_ONLY = 1004

# Duplicate identifier detected.
DUPLICATE_IDENTIFIER = 1005

# Datafile unreadable.
DATAFILE_UNREADABLE = 1006

# Datafile empty.
DATAFILE_EMPTY = 1007

# Error occurred during WAL log file recovery.
RECOVERY = 1008

# Required datafile statistics object not found.
DATAFILE_STATISTICS_NOT_FOUND = 1009

####################################
# External ArangoDB Storage Errors #
####################################

# Datafile corrupted.
CORRUPTED_DATAFILE = 1100

# Parameter file corrupted or cannot be read.
ILLEGAL_PARAMETER_FILE = 1101

# Collection contains one or more corrupted datafiles.
CORRUPTED_COLLECTION = 1102

# System call mmap failed.
MMAP_FAILED = 1103

# Filesystem full.
FILESYSTEM_FULL = 1104

# Cannot create journal.
NO_JOURNAL = 1105

# Datafile of the same name already exists.
DATAFILE_ALREADY_EXISTS = 1106

# Database directory locked by another process.
DATADIR_LOCKED = 1107

# Directory of the same name already exists.
COLLECTION_DIRECTORY_ALREADY_EXISTS = 1108

# System call msync failed.
MSYNC_FAILED = 1109

# Cannot lock the database directory on startup.
DATADIR_UNLOCKABLE = 1110

# Server waited too long for the datafile to be synced to disk.
SYNC_TIMEOUT = 1111

###################################
# General ArangoDB Storage Errors #
###################################

# Conflict detected while updating or deleting a document.
CONFLICT = 1200

# Database directory invalid.
DATADIR_INVALID = 1201

# Unknown document identifier or handle.
DOCUMENT_NOT_FOUND = 1202

# Collection with given identifier or name unknown.
DATA_SOURCE_NOT_FOUND = 1203

# Missing collection parameter.
COLLECTION_PARAMETER_MISSING = 1204

# Invalid document handle.
DOCUMENT_HANDLE_BAD = 1205

# Maximal journal size too small.
MAXIMAL_SIZE_TOO_SMALL = 1206

# Duplicate name detected.
DUPLICATE_NAME = 1207

# Illegal name detected.
ILLEGAL_NAME = 1208

# No suitable index for query.
NO_INDEX = 1209

# Unique constraint violation.
UNIQUE_CONSTRAINT_VIOLATED = 1210

# Index with unknown identifier.
INDEX_NOT_FOUND = 1212

# Cross-collection requested.
CROSS_COLLECTION_REQUEST = 1213

# Index handle corrupted.
INDEX_HANDLE_BAD = 1214

# Document too large to fit into any datafile.
DOCUMENT_TOO_LARGE = 1216

# Collection must be unloaded.
COLLECTION_NOT_UNLOADED = 1217

# Invalid collection type.
COLLECTION_TYPE_INVALID = 1218

# Failed to parse an attribute name definition.
ATTRIBUTE_PARSER_FAILED = 1220

# Corrupted document key.
DOCUMENT_KEY_BAD = 1221

# User-defined document key supplied for collections with auto key generation.
DOCUMENT_KEY_UNEXPECTED = 1222

# Database directory not writable for current user.
DATADIR_NOT_WRITABLE = 1224

# Key generator out of keys.
OUT_OF_KEYS = 1225

# Document key missing.
DOCUMENT_KEY_MISSING = 1226

# There was an attempt to create a document of invalid type.
DOCUMENT_TYPE_INVALID = 1227

# Non-existing database accessed.
DATABASE_NOT_FOUND = 1228

# Invalid database used.
DATABASE_NAME_INVALID = 1229

# Operation requested in non-system database.
USE_SYSTEM_DATABASE = 1230

# Invalid key generator.
INVALID_KEY_GENERATOR = 1232

# Undefined or invalid "_from" or "_to" values in an edge document.
INVALID_EDGE_ATTRIBUTE = 1233

# Cannot create index.
INDEX_CREATION_FAILED = 1235

# Server is write-throttled and a write operation waited too long.
WRITE_THROTTLE_TIMEOUT = 1236

# Collection type mismatch.
COLLECTION_TYPE_MISMATCH = 1237

# Collection accessed but not yet loaded.
COLLECTION_NOT_LOADED = 1238

# Document revision corrupt or missing.
DOCUMENT_REV_BAD = 1239

# Read cannot be completed by storage engine.
INCOMPLETE_READ = 1240

###################################
# Checked ArangoDB Storage Errors #
###################################

# Datafile full.
DATAFILE_FULL = 1300

# Server database directory empty.
EMPTY_DATADIR = 1301

# Operation needs to be retried.
TRY_AGAIN = 1302

# Storage engine busy.
BUSY = 1303

# Datafile merge in progress and the operation cannot be completed.
MERGE_IN_PROGRESS = 1304

# Storage engine encountered an I/O error.
IO_ERROR = 1305

###############################
# ArangoDB Replication Errors #
###############################

# Replication applier received no (or incomplete) response from master.
REPLICATION_NO_RESPONSE = 1400

# Replication applier received an invalid response from master.
REPLICATION_INVALID_RESPONSE = 1401

# Replication applier received a server error from master.
REPLICATION_MASTER_ERROR = 1402

# Replication applier tried to connect to master with incompatible version.
REPLICATION_MASTER_INCOMPATIBLE = 1403

# Replication applier connected to a different master than before.
REPLICATION_MASTER_CHANGE = 1404

# Replication applier was asked to connect to itself for replication.
REPLICATION_LOOP = 1405

# Unexpected marker found in replication log stream.
REPLICATION_UNEXPECTED_MARKER = 1406

# Found invalid replication applier state file.
REPLICATION_INVALID_APPLIER_STATE = 1407

# Found unexpected transaction ID.
REPLICATION_UNEXPECTED_TRANSACTION = 1408

# Invalid replication applier configuration.
REPLICATION_INVALID_APPLIER_CONFIGURATION = 1410

# Operation attempted while replication applier is running.
REPLICATION_RUNNING = 1411

# Replication applier stopped by user.
REPLICATION_APPLIER_STOPPED = 1412

# Replication applier started without a known start tick value.
REPLICATION_NO_START_TICK = 1413

# Replication applier started without a known start tick value.
REPLICATION_START_TICK_NOT_PRESENT = 1414

# Newborn follower submits a wrong checksum.
REPLICATION_WRONG_CHECKSUM = 1416

# Shard is not empty and follower tries a shortcut.
REPLICATION_SHARD_NONEMPTY = 1417

###########################
# ArangoDB Cluster Errors #
###########################

# Raised on some occasions when one server gets a request from another.
CLUSTER_SERVER_UNKNOWN = 1449

# Coordinator cannot create a collection as the collection ID already exists.
CLUSTER_COLLECTION_ID_EXISTS = 1453

# Coordinator cannot create an entry for a new collection in Plan hierarchy.
CLUSTER_COULD_NOT_CREATE_COLLECTION_IN_PLAN = 1454

# Coordinator sees DBServer issues when creating shards for a new collection.
CLUSTER_COULD_NOT_CREATE_COLLECTION = 1456

# Coordinator runs into a timeout for some cluster wide operation.
CLUSTER_TIMEOUT = 1457

# Coordinator cannot remove an entry for a collection in Plan hierarchy.
CLUSTER_COULD_NOT_REMOVE_COLLECTION_IN_PLAN = 1458

# Coordinator cannot remove an entry for a collection in Current hierarchy.
CLUSTER_COULD_NOT_REMOVE_COLLECTION_IN_CURRENT = 1459

# Coordinator cannot create an entry for a new database in the Plan hierarchy.
CLUSTER_COULD_NOT_CREATE_DATABASE_IN_PLAN = 1460

# Coordinator sees DBServer issues when creating databases for a new cluster.
CLUSTER_COULD_NOT_CREATE_DATABASE = 1461

# Coordinator cannot remove an entry for a database in the Plan hierarchy.
CLUSTER_COULD_NOT_REMOVE_DATABASE_IN_PLAN = 1462

# Coordinator cannot remove an entry for a database in the Current hierarchy.
CLUSTER_COULD_NOT_REMOVE_DATABASE_IN_CURRENT = 1463

# Coordinator cannot determine the shard responsible for a given document.
CLUSTER_SHARD_GONE = 1464

# Coordinator loses HTTP connection to a DBServer while transferring data.
CLUSTER_CONNECTION_LOST = 1465

# "_key" attribute specified in sharded collection which uses not only "_key"
# as sharding attribute.
CLUSTER_MUST_NOT_SPECIFY_KEY = 1466

# Coordinator gets conflicting results from different shards.
CLUSTER_GOT_CONTRADICTING_ANSWERS = 1467

# Coordinator tries to find out the shard responsible for a partial document.
CLUSTER_NOT_ALL_SHARDING_ATTRIBUTES_GIVEN = 1468

# Not allowed to update the value of a shard attribute.
CLUSTER_MUST_NOT_CHANGE_SHARDING_ATTRIBUTES = 1469

# Operation not supported in sharded collection.
CLUSTER_UNSUPPORTED = 1470

# Operation is coordinator-only.
CLUSTER_ONLY_ON_COORDINATOR = 1471

# Coordinator or DBServer cannot read the Plan.
CLUSTER_READING_PLAN_AGENCY = 1472

# Coordinator cannot truncate all shards of a cluster collection.
CLUSTER_COULD_NOT_TRUNCATE_COLLECTION = 1473

# Internal communication of the cluster for AQL produces an error.
CLUSTER_AQL_COMMUNICATION = 1474

# Operation is DBServer-only.
CLUSTER_ONLY_ON_DBSERVER = 1477

# Cannot reach a required DBServer.
CLUSTER_BACKEND_UNAVAILABLE = 1478

# Required collection out of sync during AQL execution.
CLUSTER_AQL_COLLECTION_OUT_OF_SYNC = 1481

# Coordinator cannot create an entry for a new index in Plan hierarchy.
CLUSTER_COULD_NOT_CREATE_INDEX_IN_PLAN = 1482

# Coordinator cannot remove an index from Plan hierarchy.
CLUSTER_COULD_NOT_DROP_INDEX_IN_PLAN = 1483

# One tries to create a collection with "shards_like" attribute which points
# to another collection that also has one.
CLUSTER_CHAIN_OF_DISTRIBUTESHARDSLIKE = 1484

# One tries to drop a collection to which another collection points with its
# "shard_like" attribute.
CLUSTER_MUST_NOT_DROP_COLL_OTHER_DISTRIBUTESHARDSLIKE = 1485

# One tries to create a collection which points to an unknown collection in its
# "shard_like" attribute.
CLUSTER_UNKNOWN_DISTRIBUTESHARDSLIKE = 1486

# One tries to create a collection with a "replication_factor" greater than the
# available number of DBServers.
CLUSTER_INSUFFICIENT_DBSERVERS = 1487

# Cannot drop follower.
CLUSTER_COULD_NOT_DROP_FOLLOWER = 1488

# Replication operation refused by a shard leader.
CLUSTER_SHARD_LEADER_REFUSES_REPLICATION = 1489

# Non-replication operation refused by a shard follower.
CLUSTER_SHARD_FOLLOWER_REFUSES_OPERATION = 1490

# Shard leader resigned in the meantime.
CLUSTER_SHARD_LEADER_RESIGNED = 1491

# Agency operation failed after various retries.
CLUSTER_AGENCY_COMMUNICATION_FAILED = 1492

# Servers currently competing for leadership.
CLUSTER_LEADERSHIP_CHALLENGE_ONGOING = 1495

# Operation sent to a non-leading server.
CLUSTER_NOT_LEADER = 1496

# Coordinator cannot create an entry for a new view in Plan hierarchy.
CLUSTER_COULD_NOT_CREATE_VIEW_IN_PLAN = 1497

# Coordinator tries to create a view and the ID already exists.
CLUSTER_VIEW_ID_EXISTS = 1498

# Coordinator cannot drop a collection entry in Plan hierarchy.
CLUSTER_COULD_NOT_DROP_COLLECTION = 1499

#########################
# ArangoDB Query Errors #
#########################

# Running query killed by an explicit admin command.
QUERY_KILLED = 1500

# Parsed query syntactically invalid.
QUERY_PARSE = 1501

# Empty query specified.
QUERY_EMPTY = 1502

# Runtime error caused by query.
QUERY_SCRIPT = 1503

# Number out of range.
QUERY_NUMBER_OUT_OF_RANGE = 1504

# Geo index coordinate invalid or out of range.
QUERY_INVALID_GEO_VALUE = 1505

# Invalid variable name.
QUERY_VARIABLE_NAME_INVALID = 1510

# Variable redeclared in a query.
QUERY_VARIABLE_REDECLARED = 1511

# Variable name unknown or undefined.
QUERY_VARIABLE_NAME_UNKNOWN = 1512

# Cannot acquire lock on collection.
QUERY_COLLECTION_LOCK_FAILED = 1521

# Too many collections or shards in a query.
QUERY_TOO_MANY_COLLECTIONS = 1522

# Document attribute redeclared.
QUERY_DOCUMENT_ATTRIBUTE_REDECLARED = 1530

# Undefined function called.
QUERY_FUNCTION_NAME_UNKNOWN = 1540

# Argument number mismatch.
QUERY_FUNCTION_ARGUMENT_NUMBER_MISMATCH = 1541

# Argument type mismatch.
QUERY_FUNCTION_ARGUMENT_TYPE_MISMATCH = 1542

# Invalid regex.
QUERY_INVALID_REGEX = 1543

# Invalid bind parameters.
QUERY_BIND_PARAMETERS_INVALID = 1550

# Bind parameter missing.
QUERY_BIND_PARAMETER_MISSING = 1551

# Bind parameter undeclared.
QUERY_BIND_PARAMETER_UNDECLARED = 1552

# Invalid bind parameter value or type.
QUERY_BIND_PARAMETER_TYPE = 1553

# Non-boolean value used in logical operation.
QUERY_INVALID_LOGICAL_VALUE = 1560

# Non-numeric value used in arithmetic operation.
QUERY_INVALID_ARITHMETIC_VALUE = 1561

# Divide by zero.
QUERY_DIVISION_BY_ZERO = 1562

# Non-list operand used when expecting an list operand.
QUERY_ARRAY_EXPECTED = 1563

# Function "FAIL()" called inside a query.
QUERY_FAIL_CALLED = 1569

# Geo restriction specified but no suitable geo index found.
QUERY_GEO_INDEX_MISSING = 1570

# Fulltext query performed on a collection without suitable fulltext index.
QUERY_FULLTEXT_INDEX_MISSING = 1571

# Cannot convert value to a date.
QUERY_INVALID_DATE_VALUE = 1572

# Query contains more than one data-modifying operation.
QUERY_MULTI_MODIFY = 1573

# Query contains an invalid aggregate expression.
QUERY_INVALID_AGGREGATE_EXPRESSION = 1574

# Query contains options that cannot be resolved at query compile time.
QUERY_COMPILE_TIME_OPTIONS = 1575

# Query contains an invalid options specification.
QUERY_EXCEPTION_OPTIONS = 1576

# Unusable index hint.
QUERY_FORCED_INDEX_HINT_UNUSABLE = 1577

# Dynamic function not allowed.
QUERY_DISALLOWED_DYNAMIC_CALL = 1578

# Collection data accessed after modification.
QUERY_ACCESS_AFTER_MODIFICATION = 1579

############################
# AQL User Function Errors #
############################

# User function registered with invalid name.
QUERY_FUNCTION_INVALID_NAME = 1580

# User function registered with invalid code.
QUERY_FUNCTION_INVALID_CODE = 1581

# User function not found.
QUERY_FUNCTION_NOT_FOUND = 1582

# Runtime exception raised by query function.
QUERY_FUNCTION_RUNTIME_ERROR = 1583

#############################
# AQL Query Registry Errors #
#############################

# Query received an invalid JSON.
QUERY_BAD_JSON_PLAN = 1590

# Query ID not found.
QUERY_NOT_FOUND = 1591

# User provided expression does not evaluate to true.
QUERY_USER_ASSERT = 1593

# User provided expression does not evaluate to true.
QUERY_USER_WARN = 1594

##########################
# ArangoDB Cursor Errors #
##########################

# Cursor ID not found.
CURSOR_NOT_FOUND = 1600

# Concurrent request still using the cursor.
CURSOR_BUSY = 1601

###############################
# ArangoDB Transaction Errors #
###############################

# Wrong usage of transactions. This is an internal error.
TRANSACTION_INTERNAL = 1650

# Nested transactions.
TRANSACTION_NESTED = 1651

# Unregistered collection used in transaction.
TRANSACTION_UNREGISTERED_COLLECTION = 1652

# Disallowed operation in transaction.
TRANSACTION_DISALLOWED_OPERATION = 1653

# Transaction aborted.
TRANSACTION_ABORTED = 1654

# Transaction not found.
TRANSACTION_NOT_FOUND = 1655

##########################
# User Management Errors #
##########################

# Invalid username.
USER_INVALID_NAME = 1700

# Username already exists.
USER_DUPLICATE = 1702

# User not found.
USER_NOT_FOUND = 1703

# User authenticated by an external server.
USER_EXTERNAL = 1705

######################################
# Service Management Errors (Legacy) #
######################################

# Cannot download service from central repository.
SERVICE_DOWNLOAD_FAILED = 1752

# Service upload from the client to the ArangoDB server failed.
SERVICE_UPLOAD_FAILED = 1753

###############
# LDAP Errors #
###############

# Cannot initialize an LDAP connection.
LDAP_CANNOT_INIT = 1800

# Cannot set an LDAP option.
LDAP_CANNOT_SET_OPTION = 1801

# Cannot bind to an LDAP server.
LDAP_CANNOT_BIND = 1802

# Cannot unbind from an LDAP server.
LDAP_CANNOT_UNBIND = 1803

# Cannot search the LDAP server.
LDAP_CANNOT_SEARCH = 1804

# Cannot start a TLS LDAP session.
LDAP_CANNOT_START_TLS = 1805

# LDAP did not find any objects with the specified search query.
LDAP_FOUND_NO_OBJECTS = 1806

# LDAP found zero or more than one user.
LDAP_NOT_ONE_USER_FOUND = 1807

# LDAP user not identified.
LDAP_USER_NOT_IDENTIFIED = 1808

# Cannot distinguish a valid mode for provided LDAP configuration.
LDAP_INVALID_MODE = 1820

###############
# Task Errors #
###############

# Task created with an invalid ID.
TASK_INVALID_ID = 1850

# Task created with a duplicate ID.
TASK_DUPLICATE_ID = 1851

# Task not found.
TASK_NOT_FOUND = 1852

############################
# Graph / Traversal Errors #
############################

# Invalid name passed to the server.
GRAPH_INVALID_GRAPH = 1901

# Invalid graph name passed to the server.
GRAPH_COULD_NOT_CREATE_GRAPH = 1902

# Invalid vertex ID passed to the server.
GRAPH_INVALID_VERTEX = 1903

# Vertex could not be created.
GRAPH_COULD_NOT_CREATE_VERTEX = 1904

# Vertex could not be changed.
GRAPH_COULD_NOT_CHANGE_VERTEX = 1905

# Invalid edge ID passed to the server.
GRAPH_INVALID_EDGE = 1906

# Edge could not be created.
GRAPH_COULD_NOT_CREATE_EDGE = 1907

# Edge could not be changed.
GRAPH_COULD_NOT_CHANGE_EDGE = 1908

# Too many iterations in graph traversal.
GRAPH_TOO_MANY_ITERATIONS = 1909

# Invalid filter result returned in graph traversal.
GRAPH_INVALID_FILTER_RESULT = 1910

# Edge collection may only be used once in a edge definition.
GRAPH_COLLECTION_MULTI_USE = 1920

# Collection already used by another graph in a different edge definition.
GRAPH_COLLECTION_USE_IN_MULTI_GRAPHS = 1921

# Graph name missing.
GRAPH_CREATE_MISSING_NAME = 1922

# Edge definition malformed (must be a list of dicts).
GRAPH_CREATE_MALFORMED_EDGE_DEFINITION = 1923

# Graph not found.
GRAPH_NOT_FOUND = 1924

# Graph name already exists.
GRAPH_DUPLICATE = 1925

# Vertex collection does not exist or is not part of the graph.
GRAPH_VERTEX_COL_DOES_NOT_EXIST = 1926

# Collection not a vertex collection.
GRAPH_WRONG_COLLECTION_TYPE_VERTEX = 1927

# Vertex collection not in orphan collections of the graph.
GRAPH_NOT_IN_ORPHAN_COLLECTION = 1928

# Collection already used in an edge definition of the graph.
GRAPH_COLLECTION_USED_IN_EDGE_DEF = 1929

# Edge collection not used in any edge definition of the graph.
GRAPH_EDGE_COLLECTION_NOT_USED = 1930

# Collection "_graphs" does not exist.
GRAPH_NO_GRAPH_COLLECTION = 1932

# Invalid example array object string.
GRAPH_INVALID_EXAMPLE_ARRAY_OBJECT_STRING = 1933

# Invalid example type (must be a list or dict).
GRAPH_INVALID_EXAMPLE_ARRAY_OBJECT = 1934

# Invalid number of arguments.
GRAPH_INVALID_NUMBER_OF_ARGUMENTS = 1935

# Invalid parameter type.
GRAPH_INVALID_PARAMETER = 1936

# Invalid ID.
GRAPH_INVALID_ID = 1937

# Collection already in orphans of the graph.
GRAPH_COLLECTION_USED_IN_ORPHANS = 1938

# Edge collection does not exist or is not part of the graph.
GRAPH_EDGE_COL_DOES_NOT_EXIST = 1939

# Graph has no edge collections.
GRAPH_EMPTY = 1940

# Invalid data in "_graphs" collection.
GRAPH_INTERNAL_DATA_CORRUPT = 1941

# Edge collection already defined.
GRAPH_INTERNAL_EDGE_COLLECTION_ALREADY_SET = 1942

# Orphan list argument malformed. Must be a list of strings.
GRAPH_CREATE_MALFORMED_ORPHAN_LIST = 1943

# Collection used as a relation exists.
GRAPH_EDGE_DEFINITION_IS_DOCUMENT = 1944

# Invalid/unknown session ID passed to the server.
SESSION_UNKNOWN = 1950

# Session expired.
SESSION_EXPIRED = 1951

########################
# Simple Client Errors #
########################

# This error should not happen.
SIMPLE_CLIENT_UNKNOWN_ERROR = 2000

# Client could not connect to server.
SIMPLE_CLIENT_COULD_NOT_CONNECT = 2001

# Client could not write data.
SIMPLE_CLIENT_COULD_NOT_WRITE = 2002

# Client could not read data.
SIMPLE_CLIENT_COULD_NOT_READ = 2003

#######################
# Communicator Errors #
#######################

# Communicator request aborted.
COMMUNICATOR_REQUEST_ABORTED = 2100

# Communicator disabled.
COMMUNICATOR_DISABLED = 2101

#######################
# Internal AQL errors #
#######################

# Internal error during AQL execution.
INTERNAL_AQL = 2200

# AQL block wrote in too few output registers.
WROTE_TOO_FEW_OUTPUT_REGISTERS = 2201

# AQL block wrote in too many output registers.
WROTE_TOO_MANY_OUTPUT_REGISTERS = 2202

# AQL block wrote in an output register twice.
WROTE_OUTPUT_REGISTER_TWICE = 2203

# AQL block wrote in a register that is not its output.
WROTE_IN_WRONG_REGISTER = 2204

# AQL block did not copy its input registers.
INPUT_REGISTERS_NOT_COPIED = 2205

##########################
# Foxx Management Errors #
##########################

# Service manifest file not a well-formed JSON.
MALFORMED_MANIFEST_FILE = 3000

# Service manifest contains invalid values.
INVALID_SERVICE_MANIFEST = 3001

# Service folder or bundle does not exist on the server.
SERVICE_FILES_MISSING = 3002

# Local service bundle does not match the checksum in the database.
SERVICE_FILES_OUTDATED = 3003

# Service options contain invalid values.
INVALID_FOXX_OPTIONS = 3004

# Service mountpath contains invalid characters.
INVALID_MOUNTPOINT = 3007

# No service found at given mountpath.
SERVICE_NOT_FOUND = 3009

# Service missing configuration or dependencies.
SERVICE_NEEDS_CONFIGURATION = 3010

# Service already exists at given mountpath.
SERVICE_MOUNTPOINT_CONFLICT = 3011

# Service directory does not contain a manifest file.
SERVICE_MANIFEST_NOT_FOUND = 3012

# Service options are not well-formed JSONs.
SERVICE_OPTIONS_MALFORMED = 3013

# Source path does not match a file or directory.
SERVICE_SOURCE_NOT_FOUND = 3014

# Source path could not be resolved.
SERVICE_SOURCE_ERROR = 3015

# Unknown service script.
SERVICE_UNKNOWN_SCRIPT = 3016

# API for managing Foxx services disabled.
SERVICE_API_DISABLED = 3099

###################################
# JavaScript Module Loader Errors #
###################################

# Cannot resolve module path.
MODULE_NOT_FOUND = 3100

# Module could not be parsed because of a syntax error.
MODULE_SYNTAX_ERROR = 3101

# Failed to invoke the module in its context.
MODULE_FAILURE = 3103

#####################
# Enterprise Errors #
#####################

# Requested collection needs to be smart.
NO_SMART_COLLECTION = 4000

# Given document does not have the smart graph attribute set.
NO_SMART_GRAPH_ATTRIBUTE = 4001

# Smart collection cannot be dropped.
CANNOT_DROP_SMART_COLLECTION = 4002

# "_key" not prefixed with the value of the smart graph attribute.
KEY_MUST_BE_PREFIXED_WITH_SMART_GRAPH_ATTRIBUTE = 4003

# Given smart graph attribute is illegal and cannot be used for sharding.
ILLEGAL_SMART_GRAPH_ATTRIBUTE = 4004

# Smart graph attribute of collection does not match the attribute of graph.
SMART_GRAPH_ATTRIBUTE_MISMATCH = 4005

# Invalid smart join attribute declaration.
INVALID_SMART_JOIN_ATTRIBUTE = 4006

# Key must be prefixed with smart join attribute.
KEY_MUST_BE_PREFIXED_WITH_SMART_JOIN_ATTRIBUTE = 4007

# Document lacks required smart join attribute.
NO_SMART_JOIN_ATTRIBUTE = 4008

# Cannot update the value of the smart join attribute.
CLUSTER_MUST_NOT_CHANGE_SMART_JOIN_ATTRIBUTE = 4009

#########################
# Cluster Repair Errors #
#########################

# General error during cluster repairs.
CLUSTER_REPAIRS_FAILED = 5000

# Cluster repairs not healthy enough.
CLUSTER_REPAIRS_NOT_ENOUGH_HEALTHY = 5001

# Raised on various inconsistencies regarding the replication factor.
CLUSTER_REPAIRS_REPLICATION_FACTOR_VIOLATED = 5002

# Repaired collection has some shards without DBServers.
CLUSTER_REPAIRS_NO_DBSERVERS = 5003

# Shard in collection and its prototype in the corresponding "shard_like"
# collection have mismatching leaders.
CLUSTER_REPAIRS_MISMATCHING_LEADERS = 5004

# Shard in collection and its prototype in the corresponding "shard_like"
# collection don't have the same followers.
CLUSTER_REPAIRS_MISMATCHING_FOLLOWERS = 5005

# Repaired collection does not have "shard_like" as expected.
CLUSTER_REPAIRS_INCONSISTENT_ATTRIBUTES = 5006

# Collection and its "shard_like" prototype have unequal number of DBServers.
CLUSTER_REPAIRS_MISMATCHING_SHARDS = 5007

# Move shard job failed during cluster repairs.
CLUSTER_REPAIRS_JOB_FAILED = 5008

# Move shard job disappeared before finishing.
CLUSTER_REPAIRS_JOB_DISAPPEARED = 5009

# Agency transaction failed during either sending or executing it.
CLUSTER_REPAIRS_OPERATION_FAILED = 5010

#################
# Agency Errors #
#################

# Inform message must be an object.
AGENCY_INFORM_MUST_BE_OBJECT = 20011

# Inform message must contain a uint parameter 'term'.
AGENCY_INFORM_MUST_CONTAIN_TERM = 20012

# Inform message must contain a string parameter 'ID'.
AGENCY_INFORM_MUST_CONTAIN_ID = 20013

# Inform message must contain an array 'active'.
AGENCY_INFORM_MUST_CONTAIN_ACTIVE = 20014

# Inform message must contain an object 'pool'.
AGENCY_INFORM_MUST_CONTAIN_POOL = 20015

# Inform message must contain an object 'min ping'.
AGENCY_INFORM_MUST_CONTAIN_MIN_PING = 20016

# Inform message must contain an object 'max ping'.
AGENCY_INFORM_MUST_CONTAIN_MAX_PING = 20017

# Inform message must contain an object 'timeoutMult'.
AGENCY_INFORM_MUST_CONTAIN_TIMEOUT_MULT = 20018

# Cannot rebuild readDB or the spearHead from replicated log.
AGENCY_CANNOT_REBUILD_DBS = 20021

######################
# Supervision Errors #
######################

# General supervision failure.
SUPERVISION_GENERAL_FAILURE = 20501

#####################
# Dispatcher Errors #
#####################

# Queue is full.
QUEUE_FULL = 21003

######################
# Maintenance Errors #
######################

# Maintenance action cannot be stopped once started.
ACTION_OPERATION_UNABORTABLE = 6002



from __future__ import absolute_import, unicode_literals

import pytest
from six import string_types

from aioarangodb.errno import (
    CURSOR_NOT_FOUND,
    FORBIDDEN,
    HTTP_NOT_FOUND,
    HTTP_UNAUTHORIZED,
    DATABASE_NOT_FOUND
)
from aioarangodb.exceptions import (
    ReplicationApplierConfigError,
    ReplicationApplierConfigSetError,
    ReplicationApplierStartError,
    ReplicationApplierStateError,
    ReplicationApplierStopError,
    ReplicationClusterInventoryError,
    ReplicationDumpBatchCreateError,
    ReplicationDumpBatchDeleteError,
    ReplicationDumpBatchExtendError,
    ReplicationDumpError,
    ReplicationInventoryError,
    ReplicationLoggerFirstTickError,
    ReplicationLoggerStateError,
    ReplicationMakeSlaveError,
    ReplicationServerIDError,
    ReplicationSyncError
)
from aioarangodb.tests.helpers import assert_raises
pytestmark = pytest.mark.asyncio


async def test_replication_dump_methods(db, bad_db, col, docs, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await db.replication.create_dump_batch(ttl=1000)
    assert 'id' in result and 'last_tick' in result
    batch_id = result['id']

    with assert_raises(ReplicationDumpBatchCreateError) as err:
        await bad_db.replication.create_dump_batch(ttl=1000)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    result = await db.replication.dump(
        collection=col.name,
        batch_id=batch_id,
        lower=0,
        upper=1000000,
        chunk_size=0,
        include_system=True,
        ticks=0,
        flush=True,
        deserialize=True
    )
    assert 'content' in result
    assert 'check_more' in result

    with assert_raises(ReplicationDumpError) as err:
        await bad_db.replication.dump(
            collection=col.name,
            batch_id=batch_id
        )
    assert err.value.error_code == HTTP_UNAUTHORIZED

    assert await db.replication.extend_dump_batch(batch_id, ttl=1000) is True
    with assert_raises(ReplicationDumpBatchExtendError) as err:
        await bad_db.replication.extend_dump_batch(batch_id, ttl=1000)
    assert err.value.error_code == HTTP_UNAUTHORIZED

    assert await db.replication.delete_dump_batch(batch_id) is True
    with assert_raises(ReplicationDumpBatchDeleteError) as err:
        await db.replication.delete_dump_batch(batch_id)
    assert err.value.error_code in {HTTP_NOT_FOUND, CURSOR_NOT_FOUND}


async def test_replication_inventory(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    dump_batch = await sys_db.replication.create_dump_batch(ttl=1000)
    dump_batch_id = dump_batch['id']

    result = await sys_db.replication.inventory(
        batch_id=dump_batch_id,
        include_system=True,
        all_databases=True
    )
    assert isinstance(result, dict)
    assert 'collections' not in result
    assert 'databases' in result
    assert 'state' in result
    assert 'tick' in result

    result = await sys_db.replication.inventory(
        batch_id=dump_batch_id,
        include_system=True,
        all_databases=False
    )
    assert isinstance(result, dict)
    assert 'databases' not in result
    assert 'collections' in result
    assert 'state' in result
    assert 'tick' in result

    with assert_raises(ReplicationInventoryError) as err:
        await bad_db.replication.inventory(dump_batch_id)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    await sys_db.replication.delete_dump_batch(dump_batch_id)


async def test_replication_logger_state(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await sys_db.replication.logger_state()
    assert 'state' in result
    assert 'server' in result

    with assert_raises(ReplicationLoggerStateError) as err:
        await bad_db.replication.logger_state()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_first_tick(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await sys_db.replication.logger_first_tick()
    assert isinstance(result, string_types)

    with assert_raises(ReplicationLoggerFirstTickError) as err:
        await bad_db.replication.logger_first_tick()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_applier(sys_db, bad_db, url, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    # Test replication applier state
    state = await sys_db.replication.applier_state()
    assert 'server' in state
    assert 'state' in state

    with assert_raises(ReplicationApplierStateError) as err:
        await bad_db.replication.applier_state()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test replication get applier config
    result = await sys_db.replication.applier_config()
    assert 'verbose' in result
    assert 'incremental' in result
    assert 'include_system' in result

    with assert_raises(ReplicationApplierConfigError) as err:
        await bad_db.replication.applier_config()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test replication stop applier
    result = await sys_db.replication.stop_applier()
    assert 'server' in result
    assert 'state' in result

    with assert_raises(ReplicationApplierStopError) as err:
        await bad_db.replication.stop_applier()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test replication set applier config
    result = await sys_db.replication.set_applier_config(
        endpoint=url,
        database='_system',
        username='root',
        password='passwd',
        max_connect_retries=120,
        connect_timeout=15,
        request_timeout=615,
        chunk_size=0,
        auto_start=True,
        adaptive_polling=False,
        include_system=True,
        auto_resync=True,
        auto_resync_retries=3,
        initial_sync_max_wait_time=405,
        connection_retry_wait_time=25,
        idle_min_wait_time=2,
        idle_max_wait_time=3,
        require_from_present=False,
        verbose=True,
        restrict_type='exclude',
        restrict_collections=['students']
    )
    assert result['endpoint'] == url
    assert result['database'] == '_system'
    assert result['username'] == 'root'
    assert result['max_connect_retries'] == 120
    assert result['connect_timeout'] == 15
    assert result['request_timeout'] == 615
    assert result['chunk_size'] == 0
    assert result['auto_start'] is True
    assert result['adaptive_polling'] is False
    assert result['include_system'] is True
    assert result['auto_resync'] is True
    assert result['auto_resync_retries'] == 3
    assert result['initial_sync_max_wait_time'] == 405
    assert result['connection_retry_wait_time'] == 25
    assert result['idle_min_wait_time'] == 2
    assert result['idle_max_wait_time'] == 3
    assert result['require_from_present'] is False
    assert result['verbose'] is True
    assert result['restrict_type'] == 'exclude'
    assert result['restrict_collections'] == ['students']

    with assert_raises(ReplicationApplierConfigSetError) as err:
        await bad_db.replication.set_applier_config(url)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test replication start applier
    result = await sys_db.replication.start_applier()
    assert 'server' in result
    assert 'state' in result
    await sys_db.replication.stop_applier()

    with assert_raises(ReplicationApplierStartError) as err:
        await bad_db.replication.start_applier()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_make_slave(sys_db, bad_db, url, replication):
    if not replication:
        pytest.skip('Only tested for replication')

    await sys_db.replication.stop_applier()

    result = await sys_db.replication.make_slave(
        endpoint='tcp://192.168.1.65:8500',
        database='test',
        username='root',
        password='passwd',
        restrict_type='include',
        restrict_collections=['test'],
        include_system=False,
        max_connect_retries=5,
        connect_timeout=500,
        request_timeout=500,
        chunk_size=0,
        adaptive_polling=False,
        auto_resync=False,
        auto_resync_retries=0,
        initial_sync_max_wait_time=0,
        connection_retry_wait_time=0,
        idle_min_wait_time=0,
        idle_max_wait_time=0,
        require_from_present=False,
        verbose=True
    )
    assert 'endpoint' in result
    assert 'database' in result

    with assert_raises(ReplicationMakeSlaveError) as err:
        await bad_db.replication.make_slave(endpoint=url)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_cluster_inventory(sys_db, bad_db, cluster):
    if cluster:
        result = await sys_db.replication.cluster_inventory(include_system=True)
        assert isinstance(result, dict)

    with assert_raises(ReplicationClusterInventoryError) as err:
        await bad_db.replication.cluster_inventory(include_system=True)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_server_id(sys_db, bad_db):
    result = await sys_db.replication.server_id()
    assert isinstance(result, string_types)

    with assert_raises(ReplicationServerIDError) as err:
        await bad_db.replication.server_id()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_replication_synchronize(sys_db, bad_db, url, replication):
    if not replication:
        pytest.skip('Only tested for replication')

    result = await sys_db.replication.synchronize(
        endpoint='tcp://192.168.1.65:8500',
        database='test',
        username='root',
        password='passwd',
        include_system=False,
        incremental=False,
        restrict_type='include',
        restrict_collections=['test'],
        initial_sync_wait_time=None
    )
    assert 'collections' in result
    assert 'last_log_tick' in result

    with assert_raises(ReplicationSyncError) as err:
        await bad_db.replication.synchronize(endpoint=url)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}
